
#include "client.h"

static Bool win_want_align_right(Dock *d, Window win);

Client *
client_new(Dock *dock, Window win)
{
   Client *c;
   long icccm_mask;
   XWindowAttributes attr;
   XSizeHints	*size;
   
   c = NEW(Client);

   c->next = NULL;
   c->win = win;
   c->dock = dock;
   
   if (dock->tail_client != NULL)
      dock->tail_client->next = c;
   dock->tail_client = c;

   if (dock->head_client == NULL) dock->head_client = c;

   XGetWindowAttributes(dock->dpy, win, &attr);

   size = XAllocSizeHints();
   if ( !XGetWMNormalHints(dock->dpy, win, size, &icccm_mask) )
   {
      c->w = attr.width;
      c->h = attr.height;
   } else {
      if (size->flags & PBaseSize) {
	 c->w  = size->base_width;
	 c->h =  size->base_height;
      } else if (size->flags &  PMinSize) {
	 c->w  = size->min_width;
	 c->h = size->min_height;
      } else {
	 c->w = attr.width;
	 c->h = attr.height;
      }
   }

   c->aligned_east = False;
   if (win_want_align_right(dock, win))
   {
      c->aligned_east = True;
      DBG("wanting align east\n");
   }
   
   c->x = client_get_init_x_pos(dock,c);
   c->y = client_get_init_y_pos(dock,c);

   /* TODO UTF8 naming */
   XFetchName(dock->dpy, win, &c->name);
   if (c->name == NULL) {
       XStoreName(dock->dpy, win, "<unnamed>");
       XFetchName(dock->dpy, win, &c->name);
       if (c->name == NULL) c->name = "<unnamed>";
   }
   
   if (!XGetCommand(dock->dpy, win, &c->argv, &c->argc))
   {
     /* TODO XGetWMClientMachine(display, w, text_prop_return) */
      c->argv = NULL;
      c->argc = 0;
   }
   
   DBG("docking with %i, %i : %ix%i\n", c->x, c->y, c->w, c->h);

   client_deliver_config(dock, c);


   if (c->aligned_east)
   {
      XReparentWindow(dock->dpy, c->win, dock->rwin, c->x,c->y);
   } else {
      XReparentWindow(dock->dpy, c->win, dock->lwin, c->x,c->y);
   }
   
   return c;
}

Bool
client_get_command_str(Dock *d, Client *c, char *cmd)
{
   int i;
   if (!c->argc) return False;
   strcat(cmd, c->argv[0]);
   for(i=1;i<c->argc;i++)
   {
      strcat(cmd, " ");
      strcat(cmd, c->argv[i]);
   }
   return True;
}

int
client_get_init_x_pos(Dock *d, Client *c)
{
   Client *p;
   int x;

   if (c->aligned_east)
   {
      x = d->w - d->init_padding - c->w;
      for (p = d->head_client; p->next != NULL; p = p->next)
	 if (p->mapped && p->aligned_east)
	    x -= ( d->padding + p->w );
   } else {
      
      x = d->init_padding;
      for (p = d->head_client; p->next != NULL; p = p->next)
	 if (p->mapped && !p->aligned_east)
	    x += ( d->padding + p->w );
   }
   return x;
}

int
client_get_init_y_pos(Dock *d, Client *c)
{
   return ( (d->h - c->h) / 2 );
}

void
client_destroy(Dock *dock, Client *c)
{
   Client *p;
   int x,w;
   MenuItem *tmp = NULL;
   
   if (!c) return;
   
   x = c->x;
   w = c->w;
      
   DBG("Destroying \n");

  tmp = dock->remove_menu->items;

  while (tmp != NULL)
    {
      if (tmp->next_item
	  && (Client *)tmp->next_item->cb_data == c)
	{
	  free(tmp->next_item);
	  tmp->next_item = tmp->next_item->next_item;
	  continue;
	}
      tmp = tmp->next_item;
    }
  
  /* Remove from list */
  if (dock->head_client == c)
    {
      dock->head_client = c->next;
    } else {
      
      for (p=dock->head_client; p->next != c; p = p->next);
      if ( dock->tail_client == c )
	{
	  dock->tail_client = p;
	  p->next = NULL;
	} else {
	  p->next = c->next;
	}
    }
  

   /* Realign the docked apps */
   for (p = dock->head_client; p != NULL; p = p->next)
   {
      if (c->aligned_east)
      {
	 if (p->mapped && p->x < x && p->aligned_east)
	 {
	    p->x += (w + dock->padding);
	    XMoveWindow(dock->dpy, p->win, p->x, p->y); 
	 }

      } else {
	 if (p->mapped && p->x > x && !p->aligned_east)
	 {
	    DBG("found greater moving x from %i", p->x);
	    p->x -= (w + dock->padding);
	    DBG(" to %i \n", p->x);
	    XMoveWindow(dock->dpy, p->win, p->x, p->y); 
	 }
      }
   }
   free(c);

}

void
client_handle_configure(Dock *dock, Client *c, XConfigureRequestEvent *ev)
{
  XWindowChanges xwc;

  DBG("sending configure to %s\n", c->name);

  xwc.width  = c->w;
  xwc.height = c->h;
  xwc.x = ev->x;
  xwc.y = ev->y;

  c->x = ev->x;
  c->y = ev->y;
  
  XConfigureWindow(dock->dpy, c->win,
		   ev->value_mask, &xwc);
}

void
client_deliver_config(Dock *dock, Client *c)
{
   XConfigureEvent ce;
   
   ce.type = ConfigureNotify;
   ce.event = c->win;
   ce.window = c->win;
   
   ce.x = c->x;
   ce.y = c->y + dock->y; 	/* BROKE - to do fix  */
   
   ce.width  = c->w;
   ce.height = c->h;
   ce.border_width = 0;
   ce.above = None;
   ce.override_redirect = 0;
   
   XSendEvent(dock->dpy, c->win, False,
	      StructureNotifyMask, (XEvent *)&ce);
}


Client
*client_find(Dock *dock, Window win_to_find)
{
   Client *c = dock->head_client;
   while( c != NULL)
   {
      if (c->win == win_to_find) return c;
      c = c->next;
   }
   return NULL;
}


static Bool
win_want_align_right(Dock *d, Window win)
{
   Atom realType;
   unsigned long n;
   unsigned long extra;
   int format;
   int status;
   Atom * value = NULL;

   status = XGetWindowProperty(d->dpy, win,
			       d->atoms[ATOM_MB_DOCK_ALIGN],
			       0L, 1000000L,
			       0, XA_ATOM, &realType, &format,
			       &n, &extra, (unsigned char **) &value);
   if (status == Success)
   {
      if (realType == XA_ATOM && format == 32)
      {
	 if (value[0] == d->atoms[ATOM_MB_DOCK_ALIGN_EAST])
	    return True;
      }
   }
   return False;
}

