/*
 *  mbdock
 *
 *  A 'system tray'
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */


#include "libmbmenu.h"
#include "dock.h"
#include "client.h"
#include "msg.h"

#define PAD 4

Atom mb_align_atom;
Atom mb_align_east_atom;
Display *dpy;

static void install_signal_handlers(void);
static void fork_exec(char *cmd);


static char * slider_xpm[] = {
   "8 12 4 1",
   ". c #969696",
   "# c #323232",
   "a c #c8c8c8",
   "b c #646464",
   "aaaaaaaa",
   "a......#",
   "a.ab.ab#",
   "a..b..b#",
   "a.ab.ab#",
   "a..b..b#",
   "a.ab.ab#",
   "a..b..b#",
   "a.ab.ab#",
   "a..b..b#",
   "a......#",
   "a#######" 
   /*
  "8 12 2 1",
  "  s None c None",
  ". c #000000",
  "   ..   ",
  "   ..   ",
  "   ..   ",
  "   ..   ",
  " . .. . ",
  ".. .. ..",
  ".. .. ..",
  " . .. . ",
  "   ..   ",
  "   ..   ",
  "   ..   ",
  "   ..   "
   */
};

static void
cleanup_children(int s)
{
  DBG("DIE DIE\n");
  kill(-getpid(), 15);  /* kill every one in our process group  */
  exit(0);
}

static void install_signal_handlers(void)
{

  signal (SIGCHLD, SIG_IGN);  /* kernel can deal with zombies  */
  signal (SIGINT, cleanup_children);
  signal (SIGQUIT, cleanup_children);
  signal (SIGTERM, cleanup_children);
  //  signal (SIGSEGV, cleanup_children);

}

static int handle_xerror(Display *dpy, XErrorEvent *e)
{
  char msg[255];
  XGetErrorText(dpy, e->error_code, msg, sizeof msg);
  fprintf(stderr, "Dock X error (%#lx):\n %s (opcode: %i)\n",
	  e->resourceid, msg, e->request_code);
  return 0;
}



static void dock_paint_slider(Dock *dock)
{
  XSetWindowBackgroundPixmap(dock->dpy, dock->swin, dock->slider_pxm);
  XClearWindow(dock->dpy, dock->swin);
}

static void 
dock_update_slider_vis(Dock *d)
{
  Client *c;
  int lmax = 0, rmax = d->w;
  for (c = d->head_client; c != NULL; c = c->next)
    {
      if (c->aligned_east)
	{
	  if (c->x < rmax)
	    rmax = c->x;
	} else {
	  if ((c->x+c->w) > lmax)
	    lmax = c->x+c->w;
	}
    }

  if (lmax > d->slider_x)
    { 
      DBG("mapping slider window\n");
      XMapWindow(d->dpy, d->swin); 
      
      if (rmax > d->slider_x)
	{ 			/* Move slider into place if room  */
	  d->slider_x = lmax + 1;
	  XResizeWindow(d->dpy, d->lwin,
			d->slider_x, d->h);
	  XMoveWindow(d->dpy, d->swin,
		      d->slider_x-(SLIDER_WIDTH/2),
		      (d->h-SLIDER_HEIGHT)/2);
	}
      return;
    }

  if (rmax < d->slider_x)
    {
      DBG("mapping slider window\n");
      XMapWindow(d->dpy, d->swin); 
      if (lmax < d->slider_x)
	{
	  d->slider_x = rmax - 1;
	  XResizeWindow(d->dpy, d->lwin,
			d->slider_x, d->h);
	  XMoveWindow(d->dpy, d->swin,
		      d->slider_x-(SLIDER_WIDTH/2),
		      (d->h-SLIDER_HEIGHT)/2);
	}
      return;

	  
    }




  DBG("unmapping slider window\n");
  XUnmapWindow(d->dpy, d->swin); 
}

int
xcol_from_str(Display *display, XColor *col, const char *defstr)
{
  char *str;
  const char delim[] = ",:";
  char *token;
  XColor exact;

  if ((strchr(defstr, delim[0]) != NULL)
      || (strchr(defstr, delim[1]) != NULL) )
    {
      str = strdup(defstr);

      token = strsep (&str, delim); 
      col->red = ( atoi(token) * 65535 ) / 255; 
      token = strsep (&str, delim); 
      col->green = ( atoi(token) * 65535 ) / 255;
      token = strsep (&str, delim); 
      col->blue = ( atoi(token) * 65535 ) / 255;

      return XAllocColor(display,
			 DefaultColormap(display, DefaultScreen(display)),
			 col);
    } else {
      return XAllocNamedColor(display,
			      DefaultColormap(display, DefaultScreen(display)),
			      defstr, col, &exact);
    }
}



static void fork_exec(char *cmd)
{
  pid_t pid, mypid;
  mypid = getpid();
  pid = fork();

  switch (pid) {
  case 0:
    setpgid (0, mypid); /* set pgid to parents pid  */
    execlp("/bin/sh", "sh", "-c", cmd, NULL);
    fprintf(stderr, "exec failed, cleaning up child\n");
    exit(1);
  case -1:
    fprintf(stderr, "can't fork\n"); break;
  }
}


void dock_save_session(Dock *dock)
{
  int i;
  char sessionfile[256] = "";
  Client *c = dock->head_client;

  strcpy(sessionfile,getenv("HOME"));
  strcat(sessionfile, "/.mbdock");

  if (dock->session_lock) return;

  if ((dock->fp = fopen(sessionfile, "w")) == NULL)
    { fprintf(stderr,"Unable to create Session file \n"); return; }

  while( c != NULL)
    {
      if (c->argc)
	{
	  for(i = 0; i < c->argc; i++)
	    fprintf(dock->fp, "%s ", c->argv[i]);
	  fprintf(dock->fp, "\n");
	}
      c = c->next;
    }

  fclose(dock->fp);
}


void dock_init_session(Dock *dock)
{
  char sessionfile[256];

  strcpy(sessionfile,getenv("HOME"));
  strcat(sessionfile, "/.mbdock");
   
  if ((dock->fp = fopen(sessionfile, "r")) == NULL)
    {
      fprintf(stderr,"Session file does not exist\n");
      dock->session_lock = False;
      return;
    }

  install_signal_handlers();
   
}

Bool dock_get_session_entry(Dock *dock)
{
  if (!dock->session_lock) return False;
  //dock->session_entry = (char *)malloc(sizeof(char)*512);
  if (fgets(dock->session_entry, 512, dock->fp) == NULL)
    {
      fclose(dock->fp);
      dock->session_lock = False;
      //dock->session_entry = '\0'; //NULL;
      return False;
    }
  if (   dock->session_entry[strlen(dock->session_entry)-1] == '\n')
    dock->session_entry[strlen(dock->session_entry)-1] = '\0';
  dock->session_start_time = time(NULL);
  return True;
}

void
dock_set_bg(Dock *d, char *bg_pixmap_def)
{
  XpmAttributes xpm_attr;
  Pixmap tmp_pxm;
   
  xpm_attr.valuemask = XpmCloseness;
  xpm_attr.closeness = 40000;

  if (XpmReadFileToPixmap( d->dpy, d->root, bg_pixmap_def,
			   &tmp_pxm, NULL, &xpm_attr)
      == XpmSuccess )
    {
      XSetWindowBackgroundPixmap(d->dpy, d->lwin, tmp_pxm);
      XClearWindow(d->dpy, d->lwin);
      XSetWindowBackgroundPixmap(d->dpy, d->rwin, tmp_pxm);
      XClearWindow(d->dpy, d->rwin);
      XSetWindowBackgroundPixmap(d->dpy, d->win, tmp_pxm);
      XClearWindow(d->dpy, d->win);
      XSync(d->dpy, False);
      XFreePixmap(d->dpy, tmp_pxm);
    } else {
      fprintf(stderr,"failed to load xpm: %s\n", bg_pixmap_def);
    }
}

void
dock_set_theme_from_root_prop(Dock *dock)
{
  Atom realType;
  unsigned long n;
  unsigned long extra;
  int format;
  int status;
  char * value;
  struct stat stat_info;
  char dock_cfg[256];
  char dock_cfg_data[256];
  FILE *fp;

  DBG("mb theme property change on root caught\n");
		  
  status = XGetWindowProperty(dock->dpy, dock->root,
			      dock->atoms[ATOM_MB_THEME], 
			      0L, 512L, False,
			      AnyPropertyType, &realType,
			      &format, &n, &extra,
			      (unsigned char **) &value);
	    
  if (status != Success || value == 0
      || *value == 0 || n == 0)
    {
      DBG("no _MB_THEME set on root window");
    } else {
      strcpy(dock_cfg, value);
      strcat(dock_cfg, "/dock.cfg");
      if (stat(dock_cfg, &stat_info) != -1)
	{
	  DBG("attempting to open dock config\n");
	  chdir(value);
	  if (!(fp = fopen(dock_cfg, "rb"))) return;
	  fgets(dock_cfg_data,256,fp);
	  fclose(fp);

	  DBG("dock theme data is %s\n", dock_cfg_data);
	  if (!strncmp(dock_cfg_data, "color:", 6))
	    {
	      xcol_from_str(dock->dpy, &dock->xcol, &dock_cfg_data[6]);
	      XSetWindowBackground(dock->dpy, dock->win, dock->xcol.pixel);
	      XClearWindow(dock->dpy, dock->win);
	      XSetWindowBackground(dock->dpy, dock->lwin,dock->xcol.pixel);
	      XClearWindow(dock->dpy, dock->lwin);
	      XSetWindowBackground(dock->dpy, dock->rwin,dock->xcol.pixel);
	      XClearWindow(dock->dpy, dock->rwin);
	      XSync(dock->dpy, False);
	      free(value);
	      return;
	    }
	  if (!strncmp(dock_cfg_data, "pixmap:", 7))
	    {
	      DBG("setting background\n");
	      dock_set_bg(dock, &dock_cfg_data[7]);
	      free(value);
	      return;
	    }
	}
    }
  return;
}


void dock_popup_exec_cb(MenuItem *item)
{
  //char *cmd = strdup(info);
  Dock *d = (Dock *)item->cb_data;
  char *cmd = strdup(item->info);
  char *p = cmd;

  if (d->click_x < d->slider_x)
    { /* This method for alignment is utter CRAP. TODO figoure out
         A much improved way ... */
      while ( *p != '\0') 
	{
	  if (*p == '-' && *(p+1) != '\0' && *(p+1) == 'r')
	    { *p = ' '; *(p+1) = ' '; continue; }
	  p++;
	}
    }

  fork_exec(cmd);
  free(cmd);
}

void dock_popup_exit_cb(MenuItem *item)
{
  cleanup_children(0);
}

void dock_popup_hide_cb(MenuItem *item)
{
  Dock *dock = (Dock *)item->cb_data;
  dock_toggle_visibilty(dock);
}

void dock_popup_kill_cb(MenuItem *item)
{
  Client *c = (Client *)item->cb_data;
  /* XKillClient(c->dock->dpy, c->win); */

  XGrabServer(c->dock->dpy);
  XKillClient(c->dock->dpy, c->win);
  client_destroy(c->dock, c);
  dock_save_session(c->dock);
  dock_update_slider_vis(c->dock);
  XUngrabServer(c->dock->dpy);
}

void dock_popup_update_remove(Dock *d)
{
  /*
   Client *c = d->head_client;
   if (d->head_client != d->tail_client) 
     mbmenu_remove_menu(d->mbmenu, d->remove_menu);
   while( c != NULL)
   {
     if (c->name)
       mbmenu_add_item_to_menu(d->mbmenu, d->remove_menu, c->name, NULL, 
			  c->name, dock_popup_kill_cb, (void *)d);
     c = c->next;
   }
  */
}


void 
dock_init_popup(Dock *d)
{
  Menu *m;
  char orig_wd[256];
  DIR *dp;
  struct dirent *dir_entry;
  struct stat stat_info;
  FILE *fp;
  char buf[512];


#ifdef USE_XFT
  d->mbmenu = mbmenu_new(d->dpy, "verdana,sans-7:bold", "black", 
			 "white", 
			 "blue", 
			 1, 16, NULL, NULL, MBMENU_NO_ITEM_SORT);
#else
  d->mbmenu = mbmenu_new(d->dpy, "fixed", "black", "white", "blue", 1, 16, 
			 NULL, NULL, MBMENU_NO_ITEM_SORT);
#endif

  m = mbmenu_add_path(d->mbmenu, "Add", DOCKDATADIR "/add.xpm");

  if (getcwd(orig_wd, 255) == (char *)NULL)
    {
      printf("Cant get current directory\n");
      exit(0);
    }

  if ((dp = opendir(DOCKDATADIR)) == NULL)
    {
      fprintf(stderr, "failed to open %s\n", DOCKDATADIR);
      return;
    }
  
  chdir(DOCKDATADIR);
  while((dir_entry = readdir(dp)) != NULL)
    {
      lstat(dir_entry->d_name, &stat_info);
      if (!(S_ISDIR(stat_info.st_mode)))
	{
	  char *icon = NULL, *path = NULL, *cmd = NULL, *title = NULL;
	  fp = fopen(dir_entry->d_name, "r");
	  while (fgets(buf, 512, fp) != NULL)
	    {
	      if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1] = '\0';
	      if (!strncmp(buf,"title:", 6))
		{
		  title = malloc(sizeof(char) * (strlen(buf)-5));
		  strcpy(title, &buf[6]);
		}
	      else if (!strncmp(buf,"icon:", 5))
		{
		  icon = malloc(sizeof(char) * (strlen(buf)-4));
		  strcpy(icon, &buf[5]);
		}
	      else if (!strncmp(buf,"path:", 5))
		{
		  path = malloc(sizeof(char) * (strlen(buf)-4));
		  strcpy(path, &buf[5]);

		}
	      else if (!strncmp(buf,"cmd:", 4))
		{
		  cmd = malloc(sizeof(char) * (strlen(buf)-3));
		  strcpy(cmd, &buf[4]);
		}
	    }
	  if (path && title)
	    {
	      m = mbmenu_add_path(d->mbmenu, path, NULL);
	      mbmenu_add_item_to_menu(d->mbmenu, m, title, icon, cmd, 
				      dock_popup_exec_cb, (void *)d);
	    }
	  if (title) free(title);
	  if (icon) free(icon);
	  if (path) free(path);
	  if (cmd) free(cmd);
	}
    }
  chdir(orig_wd);

  d->remove_menu = mbmenu_add_path(d->mbmenu, "Remove", 
				   DOCKDATADIR "/remove.xpm"); 

  m = mbmenu_add_path(d->mbmenu, "Remove/Dock", DOCKDATADIR "/stop.xpm");
  mbmenu_add_item_to_menu(d->mbmenu, m, "Sure?", NULL, NULL, 
			  dock_popup_exit_cb, NULL); 
  mbmenu_add_item_to_menu(d->mbmenu, m, "Cancel", NULL, NULL, NULL, NULL); 

  mbmenu_add_item_to_menu(d->mbmenu, d->mbmenu->rootmenu, "Hide", 
			  DOCKDATADIR "/down.xpm", NULL , 
			  dock_popup_hide_cb, (void *)d); 
  return;
}


Dock *init_dock(int argc, char *argv[])
{
  XGCValues gv;
  XSetWindowAttributes dattr;
  XpmAttributes xpm_attr;

  char *geometry_str = NULL;
  char *color_def = NULL;
  char *bg_pixmap_def = NULL;
  char *display_name = (char *)getenv("DISPLAY");

  Dock *d;
  int i;

#ifdef USE_XFT  
  XRenderColor colortmp;
#endif

  XSetErrorHandler(handle_xerror);
   
  d = NEW(Dock);
  memset(d, sizeof(d), 0);

  d->init_padding = 4;
  d->padding = 2;
  d->use_session  = True;
  d->click_time = 250;
   
  for (i=1; argv[i]; i++) {
    char *arg = argv[i];
    if (*arg=='-') {
      switch (arg[1]) {
      case 'd' :
	display_name = argv[i+1];
	i++;
	break;
      case 'n' :
	d->use_session  = False;
	break;
      case 'p' :
	d->padding = atoi(argv[i+1]);
	i++;
	break;
      case 'c' :
	color_def = argv[i+1];
	i++;
	break;
      case 'b' :
	bg_pixmap_def = argv[i+1];
	i++;
	break;
      case 'g' :
	geometry_str = argv[i+1];
	i++;
	break;
      case 't' :
	d->click_time = atoi(argv[i+1]);
	i++;
	break;
      default:
	exit(0);
	break;
      }
    }	
  }

  if ((d->dpy = XOpenDisplay(display_name)) == NULL)
    {
      fprintf(stderr, "%s: failed to open display", argv[0]);
    }

  d->screen = DefaultScreen(d->dpy);
  d->root = RootWindow(d->dpy, d->screen);

  d->head_client = NULL;
  d->tail_client = NULL;

  d->x = 0;
  d->y = 0;
  d->h = DOCK_HEIGHT;
  d->w = DisplayWidth(d->dpy, d->screen); 
   
  d->atoms[0] = XInternAtom(d->dpy, "_NET_WM_WINDOW_TYPE"     ,False);
   
  d->atoms[1] = XInternAtom(d->dpy, "_NET_WM_WINDOW_TYPE_DOCK",False);

  d->atoms[2] = XInternAtom(d->dpy, "_NET_SYSTEM_TRAY_S0",     False);
  d->atoms[3] = XInternAtom(d->dpy, "_NET_SYSTEM_TRAY_OPCODE", False);
  d->atoms[4] = XInternAtom(d->dpy, "_XEMBED_INFO",            False);
  d->atoms[5] = XInternAtom(d->dpy, "_XEMBED",                 False);
  d->atoms[6] = XInternAtom(d->dpy, "MANAGER",                 False);

  d->atoms[7] = XInternAtom(d->dpy, "_MB_DOCK_ALIGN",          False);
  d->atoms[8] = XInternAtom(d->dpy, "_MB_DOCK_ALIGN_EAST",     False);

  d->atoms[9] = XInternAtom(d->dpy, "_NET_SYSTEM_TRAY_MESSAGE_DATA", False);
  d->atoms[10] = XInternAtom(d->dpy, "_NET_WM_WINDOW_TYPE_SPLASH", False);

  d->atoms[11] = XInternAtom(d->dpy, "WM_PROTOCOLS", False);
  d->atoms[12] = XInternAtom(d->dpy, "WM_DELETE_WINDOW", False);

  d->atoms[13] = XInternAtom(d->dpy, "_MB_THEME", False);

   
  gv.graphics_exposures = False;
  gv.function   = GXcopy;
  gv.foreground = WhitePixel(d->dpy, d->screen);

  d->gc = XCreateGC(d->dpy, d->root,
		    GCGraphicsExposures|GCFunction|GCForeground, &gv);

  if (color_def != NULL)
    xcol_from_str(d->dpy, &d->xcol,  color_def);
  else 
    xcol_from_str(d->dpy, &d->xcol,  "200:200:200");

  dattr.background_pixel = d->xcol.pixel;
   
  d->win = XCreateWindow(d->dpy, d->root,
			 d->x, d->y, 
			 d->w, d->h,
			 0,
			 CopyFromParent,
			 CopyFromParent,
			 CopyFromParent,
			 CWBackPixel,
			 &dattr);

  d->lwin = XCreateWindow(d->dpy, d->root,
			  0, 0, 
			  d->w/2, d->h,
			  0,
			  CopyFromParent,
			  CopyFromParent,
			  CopyFromParent,
			  CWBackPixel,
			  &dattr);

  d->rwin = XCreateWindow(d->dpy, d->root,
			  0, 0, 
			  d->w, d->h,
			  0,
			  CopyFromParent,
			  CopyFromParent,
			  CopyFromParent,
			  CWBackPixel,
			  &dattr);

  d->slider_x = d->w/2;
   
  d->swin = XCreateWindow(d->dpy, d->root,
			  0, 0, 
			  SLIDER_WIDTH, SLIDER_HEIGHT, 
			  0,
			  CopyFromParent,
			  CopyFromParent,
			  CopyFromParent,
			  CWBackPixel,
			  &dattr);


  xpm_attr.valuemask = XpmCloseness;
  xpm_attr.closeness = 40000;

  if (XpmCreatePixmapFromData( d->dpy, d->swin, slider_xpm,
			       &d->slider_pxm, &d->slider_mask, &xpm_attr)
      != XpmSuccess )
    {
      fprintf(stderr, "failed to get slider image\n");
      exit(1);
    }

  if (bg_pixmap_def) dock_set_bg(d, bg_pixmap_def);

   
  XReparentWindow(d->dpy, d->rwin, d->win, 0, 0);
  XReparentWindow(d->dpy, d->lwin, d->win, 0, 0);
  XReparentWindow(d->dpy, d->swin, d->win,
		  d->slider_x-(SLIDER_WIDTH/2), (d->h-SLIDER_HEIGHT)/2);

  XChangeProperty(d->dpy, d->win, d->atoms[ATOM_WM_WINDOW_TYPE], XA_ATOM, 32, 
		  PropModeReplace,
		  (unsigned char *) &d->atoms[ATOM_WM_WINDOW_TYPE_DOCK], 1);

  XUnmapWindow(d->dpy, d->swin);

  d->session_lock = True;  /* we are loading session data  */
  d->session_entry[0] = '\0'; //NULL;

  d->msg_queue_start = NULL;
  d->msg_queue_end   = NULL;
  d->msg_win = None;   

  xcol_from_str(d->dpy, &d->msg_xcol, "gold");
  d->msg_gc = XCreateGC(d->dpy, d->root,
			GCGraphicsExposures|GCFunction|GCForeground/*|GCFont*/,
			&gv);

#ifdef USE_XFT
  if ((d->msg_font = XftFontOpenName(d->dpy, d->screen,"sans,verdana-6:bold")) 
      == NULL)
    { printf("Cant open XFT font\n"); exit(0); }

  colortmp.red   = 0;
  colortmp.green = 0;
  colortmp.blue  = 0;
  colortmp.alpha = 0xffff;
  XftColorAllocValue(d->dpy,
		     DefaultVisual(d->dpy, d->screen), 
		       DefaultColormap(d->dpy, d->screen),
		       &colortmp,
		       &d->fg_xftcol);
#else
  if (!(d->msg_font = XLoadQueryFont(d->dpy, "5x7")))
    {
      printf("5x7 font not found, giving up...\n");
      exit(1);
    }
  XSetFont(d->dpy, d->msg_gc, d->msg_font->fid);
#endif

  d->msg_font_height = d->msg_font->ascent + d->msg_font->descent; 

  d->next_click_is_not_double = True;
  d->is_hidden = False;

  XSetWMProtocols(d->dpy, d->win , &d->atoms[11], 2);

  if (bg_pixmap_def == NULL && color_def == NULL)
    dock_set_theme_from_root_prop(d);

  dock_init_popup(d); 

  if (color_def != NULL)
    mbmenu_set_bgcol(d->mbmenu, &d->xcol);

  d->click_x = 0;
  d->click_y = 0;

  /* set cursor for slider */
  d->cursors[CURS_SLIDER] = XCreateFontCursor(d->dpy, XC_sb_h_double_arrow);
  XDefineCursor(d->dpy, d->swin, d->cursors[CURS_SLIDER]);

  return d;
}


int
dock_get_map_state(Dock *d, Client *c)
{
  Atom realType;
  unsigned long n;
  unsigned long extra;
  int format;
  int status;
  CARD32 * value = NULL;
  status = XGetWindowProperty(d->dpy, c->win,
			      d->atoms[ATOM_XEMBED_INFO],
			      0L, 1000000L,
			      0, XA_CARDINAL, &realType, &format,
			      &n, &extra, (unsigned char **) &value);
  if (status == Success)
    {
      if (realType == XA_CARDINAL && format == 32)
	{
	  /*
	    printf("VERSION: %i\n", value[0]);
	    printf("MAPPED:  %i\n", value[1]);
	  */
	  return (int)value[1];
	}
    }
  return -1;
}

void dock_send_xembed_message(
			      Dock *d,
			      Client *c,
			      long message, /* message opcode */
			      long detail,  /* message detail */
			      long data1,  /* message data 1 */
			      long data2  /* message data 2 */
			      ){
  XEvent ev;
  memset(&ev, 0, sizeof(ev));
  ev.xclient.type = ClientMessage;
  ev.xclient.window = c->win;
  ev.xclient.message_type = d->atoms[ATOM_XEMBED_MESSAGE];
  ev.xclient.format = 32;
  ev.xclient.data.l[0] = CurrentTime;
  ev.xclient.data.l[1] = message;
  ev.xclient.data.l[2] = detail;
  ev.xclient.data.l[3] = data1;
  ev.xclient.data.l[4] = data2;
  XSendEvent(d->dpy, c->win, False, NoEventMask, &ev);
  XSync(d->dpy, False);
}

void dock_send_manage_message( Dock *d )
{
  XEvent ev;
  memset(&ev, 0, sizeof(ev));
  ev.xclient.type = ClientMessage;
  ev.xclient.window = d->root;
  ev.xclient.message_type = d->atoms[ATOM_MANAGER];
  ev.xclient.format = 32;
  ev.xclient.data.l[0] = CurrentTime;
  ev.xclient.data.l[1] = d->atoms[ATOM_SYSTEM_TRAY_S0];
  ev.xclient.data.l[2] = d->win;

  XSendEvent(d->dpy, d->root, False, StructureNotifyMask, &ev);
  XSync(d->dpy, False);
}


void
set_prop (Window win, Atom at, long val)
{
  XChangeProperty (dpy, win, at, XA_CARDINAL, 32,
		   PropModeReplace, (unsigned char *) &val, 1);
}


void
handle_error (Display * d, XErrorEvent * ev)
{
}

static void
get_mouse_position(Dock *d, int *x, int *y)
{
  Window mouse_root, mouse_win;
  int win_x, win_y;
  unsigned int mask;

  XQueryPointer(d->dpy, d->root, &mouse_root, &mouse_win,
		x, y, &win_x, &win_y, &mask);
}

void
dock_get_slider_close_clients(Dock *d, Client *cl, Client *cr)
{
  Client *p;
  cl = NULL; cr = NULL;

  for (p = d->head_client; p != NULL; p = p->next)
    {
      XMapWindow(d->dpy, p->win);
    }
}

void
dock_toggle_visibilty(Dock *d)
{
  if (d->is_hidden)
    {
      XResizeWindow(d->dpy, d->win, d->w, DOCK_HEIGHT);
      XMapSubwindows(d->dpy, d->win);
      d->is_hidden = False;
    } else {
      XUnmapWindow(d->dpy, d->lwin);
      XUnmapWindow(d->dpy, d->rwin);
      XUnmapWindow(d->dpy, d->swin);
      XSync(d->dpy, False);
      XResizeWindow(d->dpy, d->win, d->w, 4);
      d->is_hidden = True;
    }
}

void
dock_handle_button_event(Dock *dock, XButtonEvent *e)
{
  XEvent ev;
  int done = 0;

  DBG("button click, subwin is %li\n", e->subwindow );

  if (e->window != dock->win) return;


  if (dock->is_hidden)
    {
      dock_toggle_visibilty(dock);
      return;
    }
   
  if (e->subwindow == dock->swin)
    {
      while (!done) {
	XMaskEvent(dock->dpy,
		   (ButtonPressMask|
		    ButtonReleaseMask|
		    PointerMotionMask), &ev);
      
	switch (ev.type) {
	case MotionNotify:
	  if (ev.xmotion.x > 0
	      && ev.xmotion.x < dock->w+(SLIDER_WIDTH/2))
	    {
	      dock->slider_x = ev.xmotion.x;
	      XResizeWindow(dock->dpy, dock->lwin,
			    dock->slider_x, dock->h);
	      XMoveWindow(dock->dpy, dock->swin,
			  dock->slider_x-(SLIDER_WIDTH/2),
			  (dock->h-SLIDER_HEIGHT)/2);
	    }
	  break;
	case ButtonRelease:
	  done=1;
	  break;
	}
      }
    } else {

      Time then = e->time;
      while (!done) {
	XMaskEvent(dock->dpy,ButtonReleaseMask, &ev);
	if (ev.type == ButtonRelease) done=1;
      }

      if ((ev.xbutton.time-then) >= dock->click_time)
	{
	  get_mouse_position(dock, &dock->click_x, &dock->click_y);
	  mbmenu_activate(dock->mbmenu,dock->click_x,dock->click_y);
	}
    }
}

void
dock_handle_client_message(Dock *dock, XClientMessageEvent *e)
{
  Client *c;
  DBG("woohoo, got a message\n");
  if (e->message_type
      == dock->atoms[ATOM_SYSTEM_TRAY_OPCODE])
    {
      DBG("type is of tray\n");
      switch (e->data.l[1])
	{
	case SYSTEM_TRAY_REQUEST_DOCK:
	  DBG("attempting dock\n");
	  c = client_new(dock, e->data.l[2]);
	  XMapWindow(dock->dpy, c->win);
	  //dock_popup_update_remove(dock);

	  if (c->name)
	    mbmenu_add_item_to_menu(dock->mbmenu, dock->remove_menu, 
				    c->name, NULL, 
				    c->name, dock_popup_kill_cb, 
				    (void *)c);

	  dock_update_slider_vis(dock);

	  XSync(dock->dpy, False);
	    
	  XSelectInput(dock->dpy, c->win,
		       PropertyChangeMask );
	    
	  /* tell app its docked */
	  dock_send_xembed_message(dock, c,
				   XEMBED_EMBEDDED_NOTIFY,
				   0, dock->win, 0);
	    
	  /* sent when it gets focus */
	  dock_send_xembed_message(dock, c,
				   XEMBED_WINDOW_ACTIVATE,
				   0,0,0);
	    
	  if (dock->use_session)
	    {
	      char cmd[512] = "";
	      client_get_command_str(dock,c,cmd);

	      if (dock->session_entry
		  && c->argc
		  && (!strncmp(cmd, dock->session_entry,
			       strlen(cmd) ) ))
		{
		  dock->session_entry[0] = '\0';
		} else {
		  dock_save_session(dock);
		}
	    }
	    
	  c->mapped = True;
	  break;
	case SYSTEM_TRAY_BEGIN_MESSAGE:
	  msg_new(dock, e);
	  break;
	case SYSTEM_TRAY_CANCEL_MESSAGE:
	  break;
	}
      return;
    }
  if (e->message_type
      == dock->atoms[ATOM_NET_SYSTEM_TRAY_MESSAGE_DATA])
    {
      DBG("got message data\n");
      msg_add_data(dock, e);
      return;
    }

  if (e->message_type
      == dock->atoms[ATOM_WM_DELETE_WINDOW])
    {
      DBG("BYE");
      cleanup_children(0);
    }
   
  return;
}

void
dock_handle_property_notify(Dock *dock, XPropertyEvent *e)
{
  Client *c;
  if (e->atom == dock->atoms[ATOM_XEMBED_INFO])
    {
      int i; 
      c = client_find(dock, e->window);
      DBG("got property notify\n");
      if (c != NULL)
	{
	  i = dock_get_map_state(dock, c);
	  if (i == 1)
	    {
	      XMapWindow(dock->dpy, c->win);
	      c->mapped = True;
	    }
	  if (i == 0)
	    {
	      XUnmapWindow(dock->dpy, c->win);
	      c->mapped = False;
	    }
	}
      return;
    }
  if (e->atom == dock->atoms[ATOM_MB_THEME])
    {
      dock_set_theme_from_root_prop(dock);
      mbmenu_set_bgcol(dock->mbmenu, &dock->xcol);
    }
}


int main(int argc, char *argv[])
{
  Dock *dock;
   
  XEvent an_event;

  dock = init_dock(argc, argv);

  /* Attempt to own the system tray selection  */
  if (!XGetSelectionOwner(dock->dpy, dock->atoms[ATOM_SYSTEM_TRAY_S0]))
    {
      XSetSelectionOwner(dock->dpy, dock->atoms[ATOM_SYSTEM_TRAY_S0],
			 dock->win, CurrentTime);
    } else {
      fprintf(stderr, "dock already exists\n"); exit(0);
    }

  /* Announce to any clients that are interested that we have it */
  dock_send_manage_message( dock ); 
   
  XMapWindow(dock->dpy, dock->win);
  XMapWindow(dock->dpy, dock->lwin);
  XMapWindow(dock->dpy, dock->rwin);
  /*  XMapWindow(dock->dpy, dock->swin); */

  dock_paint_slider(dock);

  if (dock->use_session) dock_init_session(dock);
   
  XSelectInput (dock->dpy, dock->win, StructureNotifyMask|ExposureMask|
		SubstructureRedirectMask|SubstructureNotifyMask|
		ButtonPressMask|ButtonReleaseMask|PointerMotionMask);
  XSelectInput (dock->dpy, dock->lwin,
		SubstructureRedirectMask|SubstructureNotifyMask);
  XSelectInput (dock->dpy, dock->rwin,
		SubstructureRedirectMask|SubstructureNotifyMask);


  XSelectInput (dock->dpy, dock->swin, 
		ExposureMask|EnterWindowMask|LeaveWindowMask);
  XSelectInput (dock->dpy, dock->root, PropertyChangeMask);

  {
    Client *c;

    while(1)
      {
	/* catch session timeouts */
	if (dock->use_session && dock->session_entry[0] != '\0')
	  {
	    if ( (time(NULL)-dock->session_start_time)
		 > SESSION_TIMEOUT)
	      {
		fprintf(stderr, "Session timeout on %s\n", 
			dock->session_entry);
		dock->session_entry[0] = '\0';
	      }
	  }

	/* launch any preexisting sessions  */
	if (dock->use_session  
	    && dock->session_entry[0] == '\0' 
	    && dock_get_session_entry(dock))
	  fork_exec(dock->session_entry);

	msg_handle_events(dock, &an_event); 
	
	while (XPending(dock->dpy))
	  {
	    XNextEvent(dock->dpy, &an_event);
	    mbmenu_handle_xevent(dock->mbmenu, &an_event);
	    switch (an_event.type)
	      {
	      case EnterNotify:

		break;
	      case LeaveNotify:

		break;

	      case ButtonPress:
		//if (an_event.xbutton.window == dock->lwin)
		dock_handle_button_event(dock, &an_event.xbutton);
		break;
	      case ClientMessage:
		dock_handle_client_message(dock, &an_event.xclient);
		break;
	      case PropertyNotify:
		dock_handle_property_notify(dock, &an_event.xproperty);
		break;
	      case MapRequest:
		DBG("got map request\n");
		break;
	      case UnmapNotify:
		DBG("event loop: unmap\n");
		/* window should unmap ... we destroy it */
		c = client_find(dock, an_event.xunmap.window);
		if (c)
		  {
		    client_destroy(dock, c);
		    /* 
		       session only updated if remove selected from menu
		       dock_save_session(dock); 

		    */
		    dock_update_slider_vis(dock);
		  }
	      case Expose:
		DBG("event loop: expose\n");
	      case DestroyNotify:
		DBG("event loop: destroy\n");
	      case ConfigureRequest:
		DBG("event loop: conf request\n");
	      case ConfigureNotify:
		 
		if (an_event.xconfigure.window == dock->win)
		  {
		    int old_w = dock->w;
		    int lwin_inc, rwin_inc, wdiff;
		    
		    Client *p;
		    
		    DBG("event loop: conf Notify\n");
		    dock->x = an_event.xconfigure.x;
		    dock->y = an_event.xconfigure.y;
		    dock->w = an_event.xconfigure.width;
		    dock->h = an_event.xconfigure.height;

		    wdiff = dock->w - old_w;
		    
		    lwin_inc = ( old_w / dock->slider_x ) * wdiff; 
		    rwin_inc = ( old_w / ( old_w * dock->slider_x ) ) * wdiff; 


		    XResizeWindow(dock->dpy, dock->rwin,
				  dock->w, dock->h);
		    
		    for (p = dock->head_client; p != NULL; p = p->next)
		      {
			if (p->aligned_east)
			  {
			    p->x += (dock->w - old_w);
			    XMoveWindow(dock->dpy, p->win, p->x, p->y);
			  }
			client_deliver_config(dock, p);
		      }

		    dock_update_slider_vis(dock);
		  } else {
		    /* -- Broke for now --
		    Client *p = client_find(dock, an_event.xconfigure.window );
		    if (p != NULL) 
		      client_handle_configure(dock, p, &an_event.xconfigure);
		    */
		    ;
		  }


	      }
	  }
	usleep(1000);
      }
  }
   
}

