#ifndef _DOCK_H_
#define _DOCK_H_


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <signal.h>
#include <sys/wait.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/xpm.h>
#include <X11/Xmd.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#ifdef USE_XFT
 #include <X11/Xft/Xft.h>
#endif


#include "libmbmenu.h"

#define NEW(OBJ) ((OBJ *)(malloc(sizeof(OBJ))))

#ifdef DEBUG
#define DBG(txt, args... ) fprintf(stderr, "DOCK-DEBUG: " txt , ##args )
#else
#define DBG(txt, args... ) /* nothing */
#endif

#define DOCK_HEIGHT 20

#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

#define ATOM_WM_WINDOW_TYPE      0
#define ATOM_WM_WINDOW_TYPE_DOCK 1
#define ATOM_WM_WINDOW_TYPE_SPLASH 10
#define ATOM_SYSTEM_TRAY_S0      2
#define ATOM_SYSTEM_TRAY_OPCODE  3
#define ATOM_XEMBED_INFO         4
#define ATOM_XEMBED_MESSAGE      5
#define ATOM_MANAGER             6
#define ATOM_MB_DOCK_ALIGN       7
#define ATOM_MB_DOCK_ALIGN_EAST  8

#define ATOM_NET_SYSTEM_TRAY_MESSAGE_DATA 9

#define ATOM_WM_PROTOCOLS 11
#define ATOM_WM_DELETE_WINDOW  12

#define ATOM_MB_THEME  13

#define XEMBED_EMBEDDED_NOTIFY  0
#define XEMBED_WINDOW_ACTIVATE  1


#define SESSION_TIMEOUT  5    /* 5 second session timeout  */
#define SLIDER_WIDTH 8
#define SLIDER_HEIGHT 12
#define DBL_CLICK_TIME 200

#define CURS_ORIG 0
#define CURS_SLIDER 1

typedef struct _client {
   
  Window win;
  char *name;
  int x;
  int y;
  int w;
  int h;
  
  Bool mapped;
  Bool aligned_east;
  
  struct _client *next;
  char **argv;
  int  argc;
  struct _dock *dock;

} Client;

typedef struct _message_queue {

   Client *sender;
   unsigned long starttime;
   int timeout; 
   int total_msg_length;
   int current_msg_length;
   int id;
   unsigned char *data;

   Bool pending;  /* Set to true if all data */ 
   
   struct _message_queue *next;

   
} MessageQueue;


typedef struct _dock {

  Display *dpy;
  MBmenu* mbmenu;
  Window win, root;

  Window rwin, lwin, swin;
  
  int slider_x;
  Pixmap slider_pxm;
  Pixmap slider_mask;
  
  GC gc;
  
  int screen;
  XColor xcol;
  
  int x,y,w,h;
  Client *head_client;
  Client *tail_client;
  
  Atom atoms[15];
  
  int init_padding;
  int padding;
  
  struct _message_queue *msg_queue_start;
  struct _message_queue *msg_queue_end;
  
  Window msg_win; /* None if no message is displayed */
  
  GC msg_gc;
  XColor msg_xcol;

#ifdef USE_XFT
   XftFont *msg_font;
   XftColor fg_xftcol;
#else
  XFontStruct *msg_font;
#endif

  int msg_font_height;
  
  Bool use_session;
  Bool session_lock;
  char session_entry[512];
  time_t session_start_time;
  /* Add session timeout */
  FILE *fp;
  
  Window last_click_window;
  Time   last_click_time;
  Bool   next_click_is_not_double;
  Bool   is_hidden;
  
  XColor bg_col;
  Pixmap bg_tile;

  Menu *remove_menu;
  int click_x; 	/* co-ords of where mouse click happened for menu  */
  int click_y;

  Time click_time;

  Cursor cursors[2];
  
} Dock;

void dock_save_session(Dock *dock);
void dock_load_session(Dock *dock);
void dock_toggle_visibilty(Dock *d);


#endif
