/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "dockbar_client.h"

Client*
dockbar_client_new(wm *w, Window win)
{
   client *c = base_client_new(w, win); 

   c->type         = dock;
   c->configure    = &dockbar_client_configure;
   c->show         = &dockbar_client_show;
   c->hide         = &dockbar_client_hide;
   c->get_coverage = &dockbar_client_get_coverage;
   c->move_resize  = &dockbar_client_move_resize;
   c->destroy      = &dockbar_client_destroy;
   
   c->frame = c->window;
   return c;
}

void
dockbar_client_get_coverage(Client *c, int *x, int *y, int *w, int *h)
{
   *x = c->x; *y = c->y - WBW(c);
   *w = c->width + WBW(c); *h = c->height + WBW(c);
}

void
dockbar_client_configure(Client *c)
{
   MBtheme *t = c->wm->mbtheme;
   c->y = c->wm->dpy_height - wm_total_dock_height(c)- c->height - (2*WBW(c));
   c->x = 0;
   dbg("DOCKBAR: y is %i", c->y);
   
   c->width  = c->wm->dpy_width - (2*WBW(c));

   dbg("DOCK: sizing as %i %i %i %i", c->x, c->y, c->width, c->height);
   
   XSetWindowBorderWidth(c->wm->dpy, c->window, WBW(c));
   XSetWindowBorder(c->wm->dpy, c->window, t->wbc->xcol.pixel);
   
   //client_set_state(c, NormalState);
}

void
dockbar_client_move_resize(Client *c)
{
   base_client_move_resize(c);
   XResizeWindow(c->wm->dpy, c->window, c->width, c->height);
   XMoveWindow(c->wm->dpy, c->window, c->x, c->y);
}

void
dockbar_client_show(Client *c) /*TODO: show and hide share common static func*/
{
   if (client_get_state(c) == NormalState) return;

   XGrabServer(c->wm->dpy);
   
   wm_restack(c->wm, c, - c->height);
   client_set_state(c, NormalState);

   XMapRaised(c->wm->dpy, c->window);
   XUngrabServer(c->wm->dpy);
}

void
dockbar_client_hide(Client *c)
{

   if (client_get_state(c) == IconicState) return;
   XGrabServer(c->wm->dpy);
   client_set_state(c, IconicState);
   wm_restack(c->wm, c, c->height-WBW(c));
   XUngrabServer(c->wm->dpy);
}

void
dockbar_client_destroy(Client *c)
{
   //dockbar_client_hide(c);
   wm_restack(c->wm, c, c->height + WBW(c) );
   base_client_destroy(c);
}

