#include "ewmh.h"

static void set_supported(Wm *w);
static void set_compliant(Wm *w);

void
ewmh_init(Wm *w)
{
   int num_desktops = 1;
   
   w->atoms[WINDOW_TYPE_TOOLBAR]
      = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE_TOOLBAR",False);
   w->atoms[WINDOW_TYPE_DOCK]
      = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE_DOCK",False);
   w->atoms[WINDOW_TYPE_DIALOG]
      = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE_DIALOG",False);
   w->atoms[WINDOW_TYPE_SPLASH]
      = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE_SPLASH",False);

   w->atoms[WINDOW_TYPE_DESKTOP]
      = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE_DESKTOP",False);
   w->atoms[WINDOW_STATE]
      = XInternAtom(w->dpy, "_NET_WM_STATE",False);
   w->atoms[WINDOW_STATE_FULLSCREEN]
      = XInternAtom(w->dpy, "_NET_WM_STATE_FULLSCREEN",False);
   w->atoms[_NET_SUPPORTED]
      = XInternAtom(w->dpy, "_NET_SUPPORTED",False);
   w->atoms[_NET_CLIENT_LIST]
      = XInternAtom(w->dpy, "_NET_CLIENT_LIST",False);
   w->atoms[_NET_NUMBER_OF_DESKTOPS]
      = XInternAtom(w->dpy, "_NET_NUMBER_OF_DESKTOPS",False);
   w->atoms[_NET_ACTIVE_WINDOW]
      = XInternAtom(w->dpy, "_NET_ACTIVE_WINDOW",False);
   w->atoms[_NET_SUPPORTING_WM_CHECK]
      = XInternAtom(w->dpy, "_NET_SUPPORTING_WM_CHECK",False);
   w->atoms[_NET_CLOSE_WINDOW]
      = XInternAtom(w->dpy, "_NET_CLOSE_WINDOW",False);
   w->atoms[_NET_WM_NAME]
      = XInternAtom(w->dpy, "_NET_WM_NAME",False);
   w->atoms[UTF8_STRING]
      = XInternAtom(w->dpy, "UTF8_STRING",False);

   w->atoms[_NET_CLIENT_LIST_STACKING]
      = XInternAtom(w->dpy, "_NET_CLIENT_LIST_STACKING",False);
   w->atoms[_NET_CURRENT_DESKTOP]
      = XInternAtom(w->dpy, "_NET_CURRENT_DESKTOP",False);

   /* set to screen space minus dock */
   w->atoms[_NET_DESKTOP_GEOMETRY]
      = XInternAtom(w->dpy, "_NET_DESKTOP_GEOMETRY",False);
   w->atoms[_NET_WORKAREA]
      = XInternAtom(w->dpy, "_NET_WORKAREA",False);

   w->atoms[_NET_CLOSE_WINDOW]
      = XInternAtom(w->dpy, "_NET_CLOSE_WINDOW",False);
   
   /* Matchbox dock only hints - probably not needed */
   w->atoms[MB_COMMAND]
      = XInternAtom(w->dpy, "_MB_COMMAND", False);
   w->atoms[MB_DOCK_ALIGN]
      = XInternAtom(w->dpy, "_MB_DOCK_ALIGN", False);
   w->atoms[MB_DOCK_ALIGN_EAST]
      = XInternAtom(w->dpy, "_MB_DOCK_ALIGN_EAST", False);
   w->atoms[MB_DOCK_ALIGN_WEST]
      = XInternAtom(w->dpy, "_MB_DOCK_ALIGN_WEST", False);

   set_compliant(w);
   set_supported(w);

   XChangeProperty(w->dpy, w->root, w->atoms[_NET_NUMBER_OF_DESKTOPS],
		   XA_CARDINAL, 32, PropModeReplace,
		   (unsigned char *)&num_desktops, 1);

   XChangeProperty(w->dpy, w->root, w->atoms[_NET_CURRENT_DESKTOP],
		   XA_CARDINAL, 32, PropModeReplace,
		   (unsigned char *)&num_desktops, 1);

   
}

int
ewhm_handle_root_message(Wm *w, XClientMessageEvent *e)
{
  /* Handle client messages _sent_ to root window */

   Client *c = NULL;
   if (e->message_type == w->atoms[_NET_ACTIVE_WINDOW])
   {
      dbg("EWMH: got active window message");
      if ((c = wm_find_client(w, e->window, WINDOW)) != NULL)
	 c->show(c);
      return 1;
   } else if (e->message_type == w->atoms[_NET_CLOSE_WINDOW]) {
	 if ((c = wm_find_client(w, e->window, WINDOW)) != NULL)
	    client_deliver_delete(c);
	 return 1;
   } else if (e->message_type == w->atoms[WINDOW_STATE]) {
	 /* toggle e->window fullscreen ? */
	 return 1;
   } else { ; }
   
   return 0;
}



void
ewmh_update(Wm *w)  /* Updates the list of clients on root */
{
   Client *c;
   Window  *wins = NULL;
   int num = 0, cnt = 0;
   CARD32 val[4];
   
   if (w->head_client)
   {
   
      START_CLIENT_LOOP(w,c);
      num++;
      END_CLIENT_LOOP(w,c);

      wins = malloc(sizeof(Window)*num);

      START_CLIENT_LOOP(w,c);
      wins[cnt++] = c->window;
      END_CLIENT_LOOP(w,c);
   }
   
   XChangeProperty(w->dpy, w->root, w->atoms[_NET_CLIENT_LIST] ,
		   XA_WINDOW, 32, PropModeReplace,
		   (unsigned char *)wins, num);

   XChangeProperty(w->dpy, w->root, w->atoms[_NET_CLIENT_LIST_STACKING] ,
		   XA_WINDOW, 32, PropModeReplace,
		   (unsigned char *)wins, num);

   val[0] = 0; 					                    /* x */
   val[1] = 0;					                    /* y */
   val[2] = w->dpy_width;			                    /* w */
   if (w->head_client)
      val[3] = w->dpy_height - wm_total_dock_height(w->head_client);/* h */
   else
      val[3] = w->dpy_height;
   
   XChangeProperty(w->dpy, w->root, w->atoms[_NET_WORKAREA],
		   XA_CARDINAL, 32, PropModeReplace, (unsigned char *)val, 4);

   XChangeProperty(w->dpy, w->root, w->atoms[_NET_DESKTOP_GEOMETRY],
		   XA_CARDINAL, 32, PropModeReplace, (unsigned char *)val, 4);
   
   if (wins)
      free(wins);
}

void ewmh_set_active(Wm *w)
{
   unsigned long  val[1];
   if (w->main_client)
      val[0] = w->main_client->window;
   else
      val[0] = None;
   
   XChangeProperty(w->dpy, w->root, w->atoms[_NET_ACTIVE_WINDOW] ,
		   XA_WINDOW, 32, PropModeReplace,
		   (unsigned char *)val, 1);
}


static void set_supported(Wm *w) /*  */
{
   Atom supported[] = {
      w->atoms[WINDOW_TYPE_TOOLBAR],
      w->atoms[WINDOW_TYPE_DOCK],
      w->atoms[WINDOW_TYPE_DIALOG],
      w->atoms[WINDOW_TYPE_DESKTOP],
      w->atoms[WINDOW_STATE],
      w->atoms[WINDOW_STATE_FULLSCREEN],
      w->atoms[_NET_SUPPORTED],
      w->atoms[_NET_CLIENT_LIST],
      w->atoms[_NET_NUMBER_OF_DESKTOPS],
      w->atoms[_NET_ACTIVE_WINDOW],
      w->atoms[_NET_SUPPORTING_WM_CHECK],
      w->atoms[_NET_CLOSE_WINDOW],
      w->atoms[_NET_CURRENT_DESKTOP],
      w->atoms[_NET_CLIENT_LIST_STACKING],
      w->atoms[_NET_CURRENT_DESKTOP],
   };

   XChangeProperty(w->dpy, w->root, w->atoms[_NET_SUPPORTED],
		   XA_ATOM, 32, PropModeReplace,
		   (unsigned char *)supported,
		   sizeof(supported)/sizeof(Atom));
}

unsigned char *
ewmh_get_utf8_name(Wm *w, Window win)
{
  Atom type;
  int format;
  long bytes_after;
  unsigned char *str;
  long n_items;

  if (XGetWindowProperty (w->dpy, win, w->atoms[_NET_WM_NAME],
                          0, 512L,
                          False, w->atoms[UTF8_STRING],
                          &type, &format, &n_items,
                          &bytes_after, (unsigned char **)&str) 
      != Success || type == None)
    return NULL;

  return str;

}

static void set_compliant(Wm *w) /* lets clients know were compliant (ish) */
{
   unsigned long  val[1];
   Window  win;
   XSetWindowAttributes attr;
   char *app_name = "matchbox";
   
   attr.override_redirect = True;
   win = XCreateWindow(w->dpy, w->root,
		       -200, -200, 5, 5, 0,
		       CopyFromParent,
		       CopyFromParent,
		       CopyFromParent,
		       CWOverrideRedirect, &attr );

   val[0] = win;
   
   XChangeProperty(w->dpy, w->root, w->atoms[_NET_SUPPORTING_WM_CHECK],
		   XA_WINDOW, 32, PropModeReplace, (unsigned char *)val,
		   1);

   XChangeProperty(w->dpy, win, w->atoms[_NET_SUPPORTING_WM_CHECK],
		   XA_WINDOW, 32, PropModeReplace,
		   (unsigned char *)val, 1);

   /* set utf8 name  */
   XChangeProperty(w->dpy, win,
		   w->atoms[_NET_WM_NAME],
		   w->atoms[UTF8_STRING],
		   8, PropModeReplace,
                   (unsigned char *)app_name, strlen(app_name)+1);

}
