/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

/*
  $Id: main_client.c,v 1.65 2002/07/11 20:53:35 mallum Exp $
*/

#include "main_client.h"


client*
main_client_new(wm *w, Window win)
{
   Client *c = base_client_new(w, win); 
   
   c->type = mainwin;
   c->reparent     = &main_client_reparent;
   c->redraw       = &main_client_redraw;
   c->button_press = &main_client_button_press;
   c->move_resize  = &main_client_move_resize;
   c->get_coverage = &main_client_get_coverage;
   c->hide         = &main_client_hide;
   c->show         = &main_client_show;
   c->configure    = &main_client_configure;
   c->destroy      = &main_client_destroy;
   c->iconize      = &main_client_iconize;

   dbg("main client new flags are %i", c->wm->flags);
   
   main_client_check_for_state_hints(c);

   main_client_check_for_single(c);
   
   return c;
}

void
main_client_check_for_state_hints(Client *c)
{
   unsigned long n;
   unsigned long extra;
   int format;
   int status;
   Atom realType;
   Atom * value = NULL;

   status = XGetWindowProperty(c->wm->dpy, c->window,
			       c->wm->atoms[WINDOW_STATE],
			       0L, 1000000L,
			       0, XA_ATOM, &realType, &format,
			       &n, &extra, (unsigned char **) &value);
   if (status == Success)
   {
      if (realType == XA_ATOM && format == 32)
      {
	 if (value[0] == c->wm->atoms[WINDOW_STATE_FULLSCREEN])
	    c->flags ^= CLIENT_FULLSCREEN_FLAG;
      }
   }
}


void
main_client_check_for_single(Client *c)
{
   if (c->wm->flags & SINGLE_FLAG)
   {
      /* There was only one client till this came along */
      c->wm->flags ^= SINGLE_FLAG; /* turn off single flag */
      if (c->wm->main_client == NULL) dbg("NO MAIN CLIENT");
      main_client_redraw(c->wm->main_client, False);
   } else if (!c->wm->main_client) /* This must be the only client*/
      c->wm->flags ^= SINGLE_FLAG; /* turn on single flag */      
}

void
main_client_configure(Client *c)
{
   int h = wm_total_toolbar_height(c);

   c->y = main_client_title_height(c);
   c->x = 0;
       
   c->width  = c->wm->dpy_width - ( 2*WBW(c) );
   c->height = c->wm->dpy_height - c->y - h - ( 2*WBW(c) );

   if ( c->flags & CLIENT_FULLSCREEN_FLAG )
      c->height = c->wm->dpy_height;
   
   c->wm->main_client = c;  /* TODO: Should this be here ?  */
}

int
main_client_title_height(Client *c)
{
   if ( (!c->wm->config->use_title)
	|| c->flags & CLIENT_FULLSCREEN_FLAG) return 0;

   if ((c->wm->flags & TITLE_HIDDEN_FLAG) && c->type == mainwin)
      return TITLE_HIDDEN_SZ + WBW(c);

   return theme_get_frame_size(c->wm->mbtheme, FRAME_MAIN ) + WBW(c);
}

void
main_client_get_coverage(Client *c, int *x, int *y, int *w, int *h)
{
   *x = 0; *y = 0;
   *w = c->width + WBW(c);
   *h = c->height + main_client_title_height(c);
}

void
main_client_reparent(Client *c)
{
    XSetWindowAttributes attr;
    MBtheme *t = c->wm->mbtheme;

    int frm_size = theme_get_frame_size(c->wm->mbtheme, FRAME_MAIN );
   
    attr.override_redirect = True;
    attr.border_pixel      = t->wbc->xcol.pixel;
    attr.background_pixel  = t->wbc->xcol.pixel;
    attr.event_mask        = ChildMask|ButtonPressMask|ExposureMask;
    
    c->frame =
       XCreateWindow(c->wm->dpy, c->wm->root, 0, 0,
		     c->width, c->height + frm_size + WBW(c), WBW(c),
		     CopyFromParent, CopyFromParent, CopyFromParent,
		     CWOverrideRedirect|CWBorderPixel|CWEventMask|CWBackPixel,
		     &attr);

    attr.background_pixel = BlackPixel(c->wm->dpy, c->wm->screen);
    
    c->title_frame =
       XCreateWindow(c->wm->dpy, c->frame, 0, 0, c->width, frm_size, 0,
		     CopyFromParent, CopyFromParent, CopyFromParent,
		     CWOverrideRedirect|CWBackPixel|CWEventMask, &attr);
    
    
    
   XSetWindowBorderWidth(c->wm->dpy, c->window, 0);
   XAddToSaveSet(c->wm->dpy, c->window);
   XSelectInput(c->wm->dpy, c->window,
		ButtonPressMask|ColormapChangeMask|PropertyChangeMask);
   
   XReparentWindow(c->wm->dpy, c->window, c->frame, 0, frm_size + WBW(c));
}

void
main_client_move_resize(Client *c)
{
   base_client_move_resize(c);
   XResizeWindow(c->wm->dpy, c->window, c->width, c->height);
   XResizeWindow(c->wm->dpy, c->title_frame, c->width,
		 theme_get_frame_size(c->wm->mbtheme, FRAME_MAIN ));
   XMoveResizeWindow(c->wm->dpy, c->frame, c->x,
		     c->y - + main_client_title_height(c), c->width,
		     c->height + main_client_title_height(c));
}

/* redraws the frame */
void
main_client_redraw(client *c, Bool use_cache)
{

   if (!c->wm->config->use_title) return;

   if (c->wm->flags & TITLE_HIDDEN_FLAG)
   {
      XUnmapWindow(c->wm->dpy, c->title_frame);
      return;
   }

   if (use_cache && c->backing != None) return ;
   
   if (c->backing == (Pixmap)NULL)
      client_init_backing(c, c->width, c->height);

   theme_paint( c->wm->mbtheme, c, FRAME_MAIN );
   theme_paint_button(c->wm->mbtheme,c, ACTION_CLOSE, INACTIVE, FRAME_MAIN );
   theme_paint_button(c->wm->mbtheme,c, ACTION_HIDE, INACTIVE, FRAME_MAIN );
   
   if (!(c->wm->flags & SINGLE_FLAG))
   {
      dbg("painting next / prev buttons");
      theme_paint_button(c->wm->mbtheme,c, ACTION_MENU, INACTIVE, FRAME_MAIN );
      theme_paint_button(c->wm->mbtheme,c,ACTION_NEXT, INACTIVE,FRAME_MAIN );
      theme_paint_button(c->wm->mbtheme,c,ACTION_PREV, INACTIVE,FRAME_MAIN );
   } else {
      dbg("maybe destroying next / prev buttons");
      if (c->button_wins[ACTION_NEXT] != None)
      {
	 dbg("destroyed 1");
	 XDestroyWindow(c->wm->dpy, c->button_wins[ACTION_NEXT]);
	 c->button_wins[ACTION_NEXT] = None;
      }
      if (c->button_wins[ACTION_PREV] != None)
      {
	 dbg("destroyed 2");
	 XDestroyWindow(c->wm->dpy, c->button_wins[ACTION_PREV]);
	 c->button_wins[ACTION_PREV] = None;
      }
      if (c->button_wins[ACTION_MENU] != None)
      {
	 dbg("destroyed 2");
	 XDestroyWindow(c->wm->dpy, c->button_wins[ACTION_MENU]);
	 c->button_wins[ACTION_MENU] = None;
      }
   }
   
   XSetWindowBackgroundPixmap(c->wm->dpy, c->title_frame, c->backing);
   XClearWindow(c->wm->dpy, c->title_frame);

}

void main_client_button_press(client *c, XButtonEvent *e)
{
   Client *p;
   
   if (!c->wm->config->use_title) return;

   if (c->wm->flags & TITLE_HIDDEN_FLAG)
   {
      main_client_toggle_title_bar(c);
      XMapWindow(c->wm->dpy, c->title_frame);
      XMapSubwindows(c->wm->dpy, c->title_frame);
      return;
   }

   switch (client_do_button_ops(c, FRAME_MAIN, e))
   {
      case ACTION_CLOSE:
	 client_deliver_delete(c);
	 break;
      case ACTION_NEXT:
	 p = client_get_next(c, mainwin);
	 if (p) { c->hide(c); p->show(p); }
	 break;
      case ACTION_PREV:
	 p = client_get_prev(c, mainwin);
	 if (p) { c->hide(c); p->show(p); }
	 break;
      case ACTION_MENU:
	 select_client_new(c->wm);
	 break;
      case ACTION_HIDE:
	 main_client_toggle_title_bar(c);
	 break;
      case -1: 		 /* Cancelled  */
	 break;
      case 0:
	 /* Not on button */
	 break;
   }
}

void
main_client_toggle_title_bar(Client *c)
{
   Client *p;
   int prev_height = main_client_title_height(c);
   c->wm->flags ^= TITLE_HIDDEN_FLAG;
   
   XGrabServer(c->wm->dpy);

   START_CLIENT_LOOP(c->wm,p);
   if (p->type == mainwin)
   {
      if (c->wm->flags & TITLE_HIDDEN_FLAG)
      {  /* hide */
	 p->height += (prev_height - TITLE_HIDDEN_SZ -WBW(p));
	 p->y = (TITLE_HIDDEN_SZ + WBW(p));
      } else {
	 /* show */
	 p->y = main_client_title_height(p);
	 p->height -= (main_client_title_height(p) - TITLE_HIDDEN_SZ - WBW(p));
      }
      p->move_resize(p);
      XMoveWindow(c->wm->dpy, p->window, p->x, p->y);
      p->redraw(p, False);
   }
   END_CLIENT_LOOP(c->wm,p);
   
   XUngrabServer(c->wm->dpy);
}

void
main_client_hide(Client *c)
{
   base_client_hide(c);

   /* lower window to bottom of stack */
   XLowerWindow(c->wm->dpy, c->frame);
}

void
main_client_iconize(Client *c)
{
  Client *p;
  base_client_iconize(c);
  p = client_get_prev(c, mainwin);
  if (p) { p->show(p); }

}


void
main_client_show(Client *c)
{
   dbg("main_client_show: called on %s\n", c->name);
   
   client_set_state(c, NormalState);
   
   if (c->wm->flags & DESKTOP_RAISED_FLAG)
      c->wm->flags ^= DESKTOP_RAISED_FLAG; /* desktop not raised anymore */
   
   c->wm->main_client = c;
   c->mapped = True;

   XMapRaised(c->wm->dpy, c->frame);
   XMapSubwindows(c->wm->dpy, c->frame);

   /* check input focus */
   if (client_want_focus(c))
   {
      XSetInputFocus(c->wm->dpy, c->window,
		     RevertToPointerRoot, CurrentTime);
      c->wm->focused_client = c;
   }

   /* deal with transients etc */
   base_client_show(c);
}

void
main_client_destroy(Client *c)
{
   Wm *w = c->wm;

   /* What main clients are left?, need to decide what nav buttons appear */

   if (c == w->main_client)
   {
      w->main_client = client_get_prev(c, mainwin);
      
      if(w->main_client == c)  /* Is this the only main client left? */
      {
	 w->main_client = NULL;
	 if (c->wm->flags & SINGLE_FLAG)
	    c->wm->flags ^= SINGLE_FLAG; /* turn off single flag */
	 base_client_destroy(c);
	 return;
      }
   }
   
   base_client_destroy(c); 

   if (w->main_client == client_get_prev(w->main_client, mainwin))
     {  /* Is only 1 main_client left  */
       w->flags ^= SINGLE_FLAG; /* turn on single flag */      
       main_client_redraw(w->main_client, False);
     }
   
   w->main_client->show(w->main_client); /* show previous main client */

}




