#include "mbtheme.h"

static int
xcol_from_str(Display *display, XColor *col, const char *defstr)
{
  char *str;
  const char delim[] = ",:";
  char *token;
  XColor exact;

  if ((strchr(defstr, delim[0]) != NULL)
      || (strchr(defstr, delim[1]) != NULL) )
  {
     str = strdup(defstr);

     token = strsep (&str, delim); 
     col->red = ( atoi(token) * 65535 ) / 255; 
     token = strsep (&str, delim); 
     col->green = ( atoi(token) * 65535 ) / 255;
     token = strsep (&str, delim); 
     col->blue = ( atoi(token) * 65535 ) / 255;

     return XAllocColor(display,
			DefaultColormap(display, DefaultScreen(display)),
			col);
  } else {
     return XAllocNamedColor(display,
			     DefaultColormap(display, DefaultScreen(display)),
			     defstr, col, &exact);
  }
}

int 
theme_get_text_length(MBtheme *t, Tfont *font, char *str, int cnt)
{
#ifdef USE_XFT   
   XGlyphInfo extents;
   XftTextExtents8(t->wm->dpy, font->font, (unsigned char *) str,
		   cnt, &extents);
   return extents.width;
#else
   return XTextWidth(font->font, str, cnt);
#endif
}

void
theme_paint_button(MBtheme *t, Client *c, int action,
		   int state, int frame_ref )
{
   Tframe *frm = theme_get_frame(t, frame_ref );
   frm->cur_width = c->width;


   if (frm->buttons[action] == NULL) return; /* No button for this theme */


   if (c->button_wins[action] == None)
   {
      /* create the window  */
      XSetWindowAttributes attr;
      int class = CopyFromParent;
      int w,h;
      attr.override_redirect = True;
      attr.event_mask = /*ButtonPressMask|*/ ExposureMask;

      if (frm->buttons[action]->inputonly ) {
	 class = InputOnly;
	 w = gparam_convert(frm, frm->buttons[action]->w,HORIZONTAL);
	 h = gparam_convert(frm, frm->buttons[action]->h, VERTICAL);
      } else {
	 h = frm->buttons[action]->inactive->h;
	 w = frm->buttons[action]->inactive->w;
      }
      
      c->button_wins[action] =
	 XCreateWindow(c->wm->dpy, c->title_frame,
		       gparam_convert(frm, frm->buttons[action]->x,HORIZONTAL),
		       gparam_convert(frm, frm->buttons[action]->y,VERTICAL),
		       w, h, 0, CopyFromParent, class, CopyFromParent,
		       CWOverrideRedirect|CWEventMask,
		       &attr);
   }

   if (!frm->buttons[action]->inputonly)
   {
      if (state == ACTIVE)
      {
	 XSetWindowBackgroundPixmap(c->wm->dpy, c->button_wins[action],
				    frm->buttons[action]->active->pxmp);
      } else {
	 XSetWindowBackgroundPixmap(c->wm->dpy, c->button_wins[action],
				    frm->buttons[action]->inactive->pxmp);
      }

      XClearWindow(c->wm->dpy, c->button_wins[action]);   
   }
   
   XMapWindow(c->wm->dpy, c->button_wins[action]);
}

int gparam_convert( Tframe *frm, Gparam *g, int orientation)
{
   int h=frm->size, w=frm->cur_width;  /* passed in */
   if (g->unit == pixel && g->value >= 0) return g->value;

   if (g->unit == textx) return frm->label_x + g->offset;
   if (g->unit == textw) return frm->label_w + g->offset;

   
   switch (orientation)
   {
      case VERTICAL:
	 if (g->unit == pixel)
	    return h + g->value;  /* Note value is negaive  */
	 if (g->unit == percentage)
	 {
	    int tmp = ( g->value / 100.00 ) * h ;
	    return ( g->value > 0 ? tmp : h+tmp ) + g->offset;
	 }
	 break;
      case HORIZONTAL:
	 if (g->unit == pixel)
	    return w + g->value;  /* Note value is negaive  */
	 if (g->unit == percentage)
	 {
	    int tmp = ( g->value / 100.00 ) * w ;
	    return ( g->value > 0 ? tmp : w+tmp ) + g->offset;
	 }
	 break;
   }
   return 0;
}

void
theme_scale_pixmap(MBtheme *t, Pixmap p_in, Pixmap m_in,
		   Pixmap p_ret, Pixmap m_ret )
{
   int xx, yy, x, y, w, h, d, bw;
   int xo, yo, wo, ho, ddo, bwo;
   Pixmap pix;
   XGCValues gcv;
   GC mgc = None;

   XGetGeometry (t->wm->dpy, p_in, &pix, &x, &y, &w, &h, &bw, &d);
   XGetGeometry (t->wm->dpy, p_ret,&pix, &xo, &yo, &wo, &ho, &bwo, &ddo);

   /*
   pix = XCreatePixmap (app->dpy, app->root,
			app->icon_support, app->icon_support,
			DefaultDepth(app->dpy, app->screen));
   */
   
   if (m_in != None)
   {
      gcv.subwindow_mode = IncludeInferiors;
      gcv.graphics_exposures = False;
      mgc = XCreateGC(t->wm->dpy, m_in,
		      GCGraphicsExposures|GCSubwindowMode, &gcv);
   }
   
   for (y = ho - 1; y >= 0; y--)
   {
		yy = (y * h) / ho;
		for (x = wo - 1; x >= 0; x--)
		{
		   xx = (x * w) / wo;
		   if (d != DefaultDepth(t->wm->dpy, t->wm->screen))
		      XCopyPlane (t->wm->dpy, p_in, p_ret, t->gc,
				  xx, yy, 1, 1, x, y, 1);
		   else
		      XCopyArea (t->wm->dpy, p_in, p_ret, t->gc,
				 xx, yy, 1, 1, x, y);
		   if (m_in != None)
		      XCopyArea (t->wm->dpy, m_in, m_ret, mgc,
				 xx, yy, 1, 1, x, y);
		}
   }
   
   if (m_in != None)
   {
      XFreeGC (t->wm->dpy, mgc);
   }

}

void
theme_pixmap_paint(MBtheme *t, Client *c, Tpixmap *p,
		   int x, int y, int w, int h)
{
   XGCValues gc_vals;
   gc_vals.clip_x_origin = x;
   gc_vals.clip_y_origin = y;
   XChangeGC(t->wm->dpy, t->gc,
	     GCClipXOrigin|GCClipYOrigin, &gc_vals);
   
   XSetClipMask(t->wm->dpy, t->gc, p->mask);
   XCopyArea(t->wm->dpy, p->pxmp,
	     c->backing, t->gc,
	     0, 0, w, h, x, y );

   XSetClipMask(t->wm->dpy, t->gc, None);			      
}

Tframe *
theme_get_frame(MBtheme *t, int req_id)
{
   int i;
   for(i=0; i<t->frame_cnt; i++)
      if (t->frames[i].id == req_id)
	 return &t->frames[i];
   return NULL;
}

int
theme_get_frame_size( MBtheme *t, int frame_ref )
{
   Tframe *frm = theme_get_frame(t, frame_ref );
   if (frm)
      return frm->size;
   return 0;
}

void
theme_paint( MBtheme *t, Client *c, int frame_ref )
 {
   int i;
   Tlayer *lyr;
   Tframe *frm = theme_get_frame(t, frame_ref );
   int tx=0,ty,txx,tyy,th,tw,j=0;
   frm->cur_width = c->width;

   dbg("THEME: paint start : frame id: %i \n", frame_ref);
   
   XSetForeground(t->wm->dpy, t->gc, WhitePixel(t->wm->dpy, t->wm->screen));
   XFillRectangle(t->wm->dpy, c->backing, t->gc, 0,0,frm->cur_width,frm->size);

   /* first figure out label dimentions */
   for(i=0; i < frm->layer_cnt; i++)
   {
      lyr = &frm->layers[i];
      if (lyr->type == label)
      {
	 int lw;
	 Alignment local_justify = lyr->label->justify;
	 j = strlen(c->name);

	 lw = theme_get_text_length(t, lyr->label->font, c->name, j);
	 
	 tw = gparam_convert(frm, lyr->w, HORIZONTAL);
	 dbg("frame text length is %i, space available is %i ( %i )\n", 
	     lw, tw, c->width);

	 if (lw > tw)
	 {
	    while ((lw = theme_get_text_length(t,lyr->label->font, c->name,j)) 
		   > tw && j>0)
	       j--;
	    local_justify = left; 
	 }
	 frm->label_w = lw;
	 dbg("THEME: frame label width is %i\n", lw);
	 tx = gparam_convert(frm, lyr->x, HORIZONTAL);
	 
	 switch(local_justify)
	 {
	    case left:
	       frm->label_x = tx;
	       break;
	    case right:
	       frm->label_x = tx + tw - lw;
	       break;
	    case center:
	       frm->label_x = tx + ( (tw-lw) / 2 );
	       break;
	 }
      }
   }

   for(i=0; i < frm->layer_cnt; i++)
   {
      XGCValues gc_vals;

      lyr = &frm->layers[i];
      
      switch (lyr->type)
	 {
	    case plain:
	       dbg("THEME: Drawing rect %i, %i, %i, %i\n",
			      gparam_convert(frm, lyr->x, HORIZONTAL),
			      gparam_convert(frm, lyr->y, VERTICAL),
			      gparam_convert(frm, lyr->w, HORIZONTAL),
			      gparam_convert(frm, lyr->h, VERTICAL));
	       XSetForeground(t->wm->dpy, t->gc, lyr->color->xcol.pixel);
	       XFillRectangle(t->wm->dpy, c->backing,
			      t->gc,
			      gparam_convert(frm, lyr->x, HORIZONTAL),
			      gparam_convert(frm, lyr->y, VERTICAL),
			      gparam_convert(frm, lyr->w, HORIZONTAL),
			      gparam_convert(frm, lyr->h, VERTICAL));
	       /*
		 something like;

		 xfillrect( gparam_convert(lyr->x), ... )

	       */
	       break;
	    case gradientHoriz:
	    case gradientVert:
	    case pixmap:
	       dbg("THEME: painting pixmap");
	       if (lyr->w->unit == object)
	       {
		  tw = lyr->tpxmp->w;
	       } else {
		  tw = gparam_convert(frm, lyr->w, HORIZONTAL);
	       }
	       if (lyr->h->unit == object)
	       {
		  th = lyr->tpxmp->h;
	       } else {
		  th = gparam_convert(frm, lyr->h, VERTICAL);
	       }
	       if (tw != lyr->tpxmp->w || th != lyr->tpxmp->h)
	       {
		  Pixmap tmp_p, tmp_m;
		  tmp_p = XCreatePixmap(t->wm->dpy,
					c->backing,
					tw, th,
					DefaultDepth(t->wm->dpy,
						     t->wm->screen));
		  tmp_m = XCreatePixmap(t->wm->dpy,
					c->backing,
					tw, th, 1);
		  theme_scale_pixmap(t, lyr->tpxmp->pxmp, lyr->tpxmp->mask,
				     tmp_p, tmp_m );

		  gc_vals.clip_x_origin = gparam_convert(frm, lyr->x,
							 HORIZONTAL),
		  gc_vals.clip_y_origin = gparam_convert(frm, lyr->x,
							 VERTICAL),
		  XChangeGC(t->wm->dpy, t->gc,
			    GCClipXOrigin|GCClipYOrigin, &gc_vals);
   
		  XSetClipMask(t->wm->dpy, t->gc, tmp_m);
		  
		  XCopyArea(t->wm->dpy, tmp_p,
			    c->backing, t->gc,
			    0, 0, tw, th,
			    gparam_convert(frm, lyr->x, HORIZONTAL),
			    gparam_convert(frm, lyr->y, VERTICAL)
			    );

		  XSetClipMask(t->wm->dpy, t->gc, None);
		  XFreePixmap(t->wm->dpy, tmp_p);
		  XFreePixmap(t->wm->dpy, tmp_m);
	       } else {
		  theme_pixmap_paint(t, c, lyr->tpxmp,
				     gparam_convert(frm, lyr->x, HORIZONTAL),
				     gparam_convert(frm, lyr->y, VERTICAL),
				     tw, th);
	       }
	       break;
	    case pixmapTiled:
	       /* TODO: figure out how to clip properly !! */
	       dbg("THEME: painting pixmap tiled");
	       
	       tw =  gparam_convert(frm, lyr->w, HORIZONTAL);
	       th =  gparam_convert(frm, lyr->h, VERTICAL);
	       tx =  gparam_convert(frm, lyr->x, HORIZONTAL);
	       ty =  gparam_convert(frm, lyr->y, VERTICAL);

	       dbg("THEME: painting pixmap tiled %i x %i, %i x %i",tw,th,tx,ty);
	       /* KLUDGE 
	       tmp_pxm = XCreatePixmap(t->wm->dpy, t->wm->root, tw, th,
				       DefualtDepth(t->wm->dpy,t->wm->screen));
	       */
	       {
		  /*
		  XRectangle r[1];
		  r[0].x = 0; r[0].y =0; r[0].width = tw; r[0].height = th;
		  
		  XSetClipRectangles(t->wm->dpy, t->gc, 0, 0, r, 1, Unsorted);
		  */
		  for(txx=tx; txx < (tx +tw ); txx += lyr->tpxmp->w )
		     for(tyy=ty; tyy < (ty + th ); tyy += lyr->tpxmp->h)
		     {
			/*
			if ( (txx+lyr->tpxmp->w) > (tx+tw) )
			{
			   printf("(txx+lyr->tpxmp->w) > (tx+tw) is TRUE\n");
			   printf(" %i, + %i, > %i + %i\n",
				  txx, lyr->tpxmp->w, tx, tw);

			}
			*/
			theme_pixmap_paint(t, c, lyr->tpxmp,
					   txx, tyy, 
					   ( (txx+lyr->tpxmp->w) > (tx+tw) )
					   ? (txx+lyr->tpxmp->w) - (tx-tw)
					   : lyr->tpxmp->w,
					   ( (tyy+lyr->tpxmp->h) > (ty+th) )
					   ? (ty+lyr->tpxmp->h) - (ty+th)
					   : lyr->tpxmp->h );

		     }
		  /* XSetClipMask(t->wm->dpy, t->gc, None);		  */
	       }
	       break;
	    case picture:

	       gc_vals.clip_x_origin = gparam_convert(frm, lyr->x, HORIZONTAL);
	       gc_vals.clip_y_origin = gparam_convert(frm, lyr->y, VERTICAL);
	       XChangeGC(t->wm->dpy, t->gc,
			 GCClipXOrigin|GCClipYOrigin, &gc_vals);

	       XSetClipMask(t->wm->dpy, t->gc, lyr->picture->mask);

	       XCopyArea(t->wm->dpy, lyr->picture->pxmp,
			 c->backing, t->gc,
			 0, 0, lyr->picture->w, lyr->picture->h,
			 gparam_convert(frm, lyr->x, HORIZONTAL),
			 gparam_convert(frm, lyr->y, VERTICAL));
			 
	       XSetClipMask(t->wm->dpy, t->gc, None);			      
	       break;
	    case icon:
	       dbg("THEME: painting icon");
	       if (c->icon)
	       {
		  XGCValues gc_vals;
		  tx =  gparam_convert(frm, lyr->x, HORIZONTAL);
		  ty =  gparam_convert(frm, lyr->y, VERTICAL);
		  
		  gc_vals.clip_x_origin = tx;
		  gc_vals.clip_y_origin = ty;
		  XChangeGC(t->wm->dpy, t->gc,
			    GCClipXOrigin|GCClipYOrigin, &gc_vals);
		  
		  if (c->icon_mask)
		     XSetClipMask(t->wm->dpy, t->gc, c->icon_mask);
	    
		  XCopyArea(t->wm->dpy, c->icon, c->backing, t->gc,
			    0, 0, 16, 16, tx, ty);
		  XSetClipMask(t->wm->dpy, t->gc, None);
	       }


	       break;

	    case label:
	       break;
	 }
   }

   for(i=0; i < frm->layer_cnt; i++)  /* finally text is painted */
   {
      lyr = &frm->layers[i];
      if (lyr->type == label)
      {
#ifdef USE_XFT
	 XftDrawString8(c->xftdraw, &(lyr->label->col->xftcol), 
			lyr->label->font->font,
			frm->label_x, 
			gparam_convert(frm, lyr->y, VERTICAL)
			+ lyr->label->font->font->ascent,
			(unsigned char *) c->name, j );
#else
	 XSetFont(t->wm->dpy, t->gc, lyr->label->font->font->fid);
	 XSetForeground(t->wm->dpy, t->gc, lyr->label->col->xcol.pixel);
	 XDrawString(t->wm->dpy, c->backing, t->gc,
		     frm->label_x, 
	             gparam_convert(frm, lyr->y, VERTICAL)
		     + lyr->label->font->font->ascent,
		     c->name, j);
#endif
      }
   }
   dbg("THEME: paint end\n");
}


/* ------------- MBtheme loading / initing below  */

Gparam *
gparam_new(char *def_str)
{
   Gparam *g;
   char *p = def_str;
   
   if (def_str == NULL) return NULL;

   g = (Gparam *)malloc(sizeof(Gparam));
   g->offset = 0;
   
   if (!strncmp(def_str,"object",5))
   {
        g->unit = object;
	g->value = 1;
	/* TODO pass offset */
	return g;
   }

   if (!strncmp(def_str,"labelx",6))
   {
        g->unit = textx;
	g->value = 1;
	if (def_str[6] != '\0')
	   g->offset = strtol(&def_str[6], (char **)NULL, 10);
	return g;
   }

   if (!strncmp(def_str,"labelw",6))
   {
      dbg("THEME: got a labelw !!!!!!!\n");
      g->unit = textw;
      g->value = 1;
      if (def_str[6] != '\0')
	 g->offset = strtol(&def_str[6], (char **)NULL, 10);
      return g;
   }
   
   while(*p != '%' && *p != '\0') { p++; }

   if (*p == '%')
   {
      *p = '\0';
      g->unit = percentage;
      g->value = strtol(def_str, (char **)NULL, 10);

      if (*(p+1) != '\0')
      {
	 g->offset = strtol(p+1, (char **)NULL, 10);
      }
   } else {
      g->unit = pixel;
      g->value = strtol(def_str, (char **)NULL, 10);
   }

   return g;
}

Tlayer *
tlayer_new()
{
   Tlayer *tl;
   tl = (Tlayer *)malloc(sizeof(Tlayer));
   memset(tl,sizeof(Tlayer),0);
   return tl;
}

static char *
get_attr(XMLNode *node, char *key)
{
   Params *p;
   for(p = node->attr; p != NULL; p = p->next)
   {
      if (!(strcmp(p->key, key)))
	 return p->value;
   }
   return NULL;
}

Tcolor *tcolor_new(MBtheme *t, char *defstr, char *name)
{
#ifdef USE_XFT   
   XRenderColor colortmp;
#endif

   Tcolor *c = (Tcolor *)malloc(sizeof(Tcolor));
   c->name = strdup(name);

   if (!xcol_from_str(t->wm->dpy, &c->xcol, defstr))
   {
      fprintf(stderr, "XAllocColor failed, attemping to alloc a cell");
   }

#ifdef USE_XFT      
   colortmp.red   = c->xcol.red;
   colortmp.green = c->xcol.green;
   colortmp.blue  = c->xcol.blue;
   colortmp.alpha = 0xffff;
   XftColorAllocValue(t->wm->dpy,
		       DefaultVisual(t->wm->dpy, t->wm->screen), 
		       DefaultColormap(t->wm->dpy, t->wm->screen),
		       &colortmp,
		       &c->xftcol);
#endif

   return c;
}


static int
handle_color_tag(MBtheme *theme, XMLNode *node)
{
   char *name, *defstr;

   if ( (name = get_attr(node, "id")) == NULL)
   { return 0; }
   if ( (defstr = get_attr(node, "def")) == NULL)
   { return 0; }
   
   theme->colors[theme->color_cnt] = tcolor_new(theme, defstr, name);
   theme->color_cnt++;
   return 1;
}

Tcolor *tcolor_lookup(MBtheme *theme, char *name)
{
   int i;
   if (name==NULL) return NULL;
   for(i=0; i<theme->color_cnt; i++)
      if (!strcmp(name, theme->colors[i]->name))
	 return theme->colors[i];
   return NULL;
}

/* Font */

Tfont *tfont_new(MBtheme *t, char *defstr, char *name)
{
   const char delim[] = "|";
   char *str, *token;
   Tfont *f = (Tfont *)malloc(sizeof(Tfont));

   if ((strchr(defstr, delim[0]) != NULL))
   {
      str = strdup(defstr);
      
      while( (token = strsep (&str, delim)) != NULL )
      {
#ifdef USE_XFT
	if ((f->font = XftFontOpenName(t->wm->dpy, t->wm->screen, token)) 
	    != NULL)
	  { 
	    f->name = strdup(name);
	    return f;
	  }
#else
	 if ((f->font = XLoadQueryFont(t->wm->dpy, token)) != NULL)
	 {
	    f->name = strdup(name);
	    return f;
	 }
#endif
      }
      
   } else {
#ifdef USE_XFT
	if ((f->font = XftFontOpenName(t->wm->dpy, t->wm->screen, defstr)) 
	    != NULL)
	  { 
	    f->name = strdup(name);
	    return f;
	  }
#else
      if ((f->font = XLoadQueryFont(t->wm->dpy, defstr)) != NULL)
      {
	 f->name = strdup(name);
	 return f;
      };
#endif
   }
   
   fprintf(stderr, "%s font not found, giving up...\n", defstr);
   exit(1);
}


static int
handle_font_tag(MBtheme *theme, XMLNode *node)
{
   char *name, *defstr;

   if ( (name = get_attr(node, "id")) == NULL)
   { return 0; }
   if ( (defstr = get_attr(node, "def")) == NULL)
   { return 0; }
   
   theme->fonts[theme->font_cnt] = tfont_new(theme, defstr, name);
   theme->font_cnt++;
   return 1;
}

Tfont *tfont_lookup(MBtheme *theme, char *name)
{
   int i;
   if (name==NULL) return NULL;
   for(i=0; i<theme->font_cnt; i++)
      if (!strcmp(name, theme->fonts[i]->name))
	 return theme->fonts[i];
   return NULL;
}


Tpixmap *tpixmap_new(MBtheme *t, char *filename, char *name)
{
   Tpixmap *p = (Tpixmap *)malloc(sizeof(Tpixmap));

   XpmAttributes attrib;
   
   attrib.valuemask = XpmColormap | XpmCloseness;
   attrib.closeness = 40000;
   attrib.colormap = DefaultColormap(t->wm->dpy, t->wm->screen);
   
   if (XpmReadFileToPixmap( t->wm->dpy, t->wm->root, filename,
			    &p->pxmp, &p->mask, &attrib) != XpmSuccess )
   {
      fprintf(stderr, "Failed loading image '%s'\n", filename);
      exit(1);
   }
   
   p->w = attrib.width;
   p->h = attrib.height;

   p->name = strdup(name);

   return p;
}


static int
handle_pixmap_tag(MBtheme *theme, XMLNode *node)
{
   /* becomes a pixmap */
   char *name, *filename;

   if ( (name = get_attr(node, "id")) == NULL)
   { return 0; }
   if ( (filename = get_attr(node, "filename")) == NULL)
   { return 0; }

   theme->pixmaps[theme->pixmap_cnt]
      = tpixmap_new(theme, filename, name);

   theme->pixmap_cnt++;
   
   return 1;
}

Tpixmap *tpixmap_lookup(MBtheme *theme, char *name)
{
   int i;
   for(i=0; i<theme->pixmap_cnt; i++)
   {
      if (!strcmp(name, theme->pixmaps[i]->name))
	 return theme->pixmaps[i];
   }
   return NULL;
}



Tpixmap *tpicture_new(MBtheme *t, int w, int h, Tcolor *tcol, char *name)
{
   Tpixmap *p = (Tpixmap *)malloc(sizeof(Tpixmap));

   p->pxmp = XCreatePixmap(t->wm->dpy, t->wm->root, w, h,
			   DefaultDepth(t->wm->dpy, t->wm->screen)
			   );
   p->mask = XCreatePixmap(t->wm->dpy, t->wm->root, w, h, 1);

   if (t->mask_gc == None) /* TODO MAYBE MOVE SOMEWHERE ELSE  */
   {
      XGCValues gv;
      gv.graphics_exposures = False;
      gv.function   = GXcopy;
      t->mask_gc = XCreateGC(t->wm->dpy, p->mask,
			     GCGraphicsExposures|GCFunction, &gv);
   }
   
   p->w = w;
   p->h = h;
   p->name = strdup(name);

   if (tcol == NULL)
   {
      XSetForeground(t->wm->dpy, t->gc, WhitePixel(t->wm->dpy, t->wm->screen));
   } else {
      XSetForeground(t->wm->dpy, t->gc, tcol->xcol.pixel);
   }
   XFillRectangle(t->wm->dpy, p->pxmp, t->gc, 0, 0, w, h );
   
   XSetForeground(t->wm->dpy, t->mask_gc,
		  BlackPixel(t->wm->dpy, t->wm->screen));
   XFillRectangle(t->wm->dpy, p->mask, t->mask_gc, 0, 0, w, h );
   
   return p;
}

Tpixmap *tpicture_lookup(MBtheme *theme, char *name)
{
   int i;
   for(i=0; i<theme->picture_cnt; i++)
      if (!strcmp(name, theme->pictures[i]->name))
	 return theme->pictures[i];
   return NULL;
}

static int
handle_picture_tag(MBtheme *theme, XMLNode *node)
{
   /* becomes a pixmap */
   int pic_w, pic_h;
   char *name;
   char *col;
   Tcolor *tcol = NULL;
   Nlist *n;
   XMLNode *inode;

   GET_INT_ATTR(node,"width" ,pic_w);
   GET_INT_ATTR(node,"height" ,pic_h);

   if ((col = get_attr(node, "bgcolor")) != NULL)
   {
      if ((tcol = tcolor_lookup(theme, col)) == NULL)
	    return 0;
   }
      
   if ( (name = get_attr(node, "id")) == NULL)
   { return 0; }
   
   theme->pictures[theme->picture_cnt]
      = tpicture_new(theme, pic_w, pic_h, tcol, name);
   
   for(n = node->kids; n != NULL; n = n->next)
   {
      inode = n->data;
      if (!(strcmp("line", inode->tag)))
      {
	 int x1,y1,x2,y2,w;
	 char *color;
	 Tcolor *tcol;

	 GET_INT_ATTR(inode,"x1", x1);
	 GET_INT_ATTR(inode,"x2", x2);
	 GET_INT_ATTR(inode,"y1", y1);
	 GET_INT_ATTR(inode,"y2", y2);
	 w = (get_attr(inode, "width") ? atoi(get_attr(inode, "width")) : 1);
	 if ( (color = get_attr(inode, "color")) == NULL)
	 { return 0; }
	 if ((tcol = tcolor_lookup(theme, color)) == NULL)
	    return 0;

	 XSetForeground(theme->wm->dpy, theme->gc, tcol->xcol.pixel);
	 XSetLineAttributes(theme->wm->dpy, theme->gc, w, LineSolid,
			    CapNotLast, JoinMiter);
	 XDrawLine(theme->wm->dpy, theme->pictures[theme->picture_cnt]->pxmp,
		   theme->gc, x1, y1, x2, y2);

	 XSetForeground(theme->wm->dpy, theme->mask_gc,
			WhitePixel(theme->wm->dpy, theme->wm->screen));
	 XSetLineAttributes(theme->wm->dpy, theme->mask_gc, w, LineSolid,
			    CapNotLast, JoinBevel);
	 XDrawLine(theme->wm->dpy, theme->pictures[theme->picture_cnt]->mask,
		   theme->mask_gc, x1, y1, x2, y2);

	 

	 continue;
      }
      if (!(strcmp("circle", inode->tag)))
      {
	 int x,y,r,lw;
	 char *color;
	 Tcolor *tcol;
	 Bool is_filled = False;
	
	 GET_INT_ATTR(inode,"x", x);
	 GET_INT_ATTR(inode,"y", y);
	 GET_INT_ATTR(inode,"r", r);

	 /* TODO: move below code into seperate func */
	 lw = (get_attr(inode, "width") ? atoi(get_attr(inode, "width")) : 1);
	 if ( (color = get_attr(inode, "color")) == NULL)
	 { return 0; }
	 if ((tcol = tcolor_lookup(theme, color)) == NULL)
	    return 0;
	 if (get_attr(inode, "options")
	     && (!strcmp("filled", get_attr(inode, "options"))))
	    is_filled = True;
	 
	 dbg("drawing circle");

	 XSetForeground(theme->wm->dpy, theme->gc, tcol->xcol.pixel);
	 XSetLineAttributes(theme->wm->dpy, theme->gc, lw, LineSolid,
			    CapNotLast, JoinMiter);
	 if (is_filled)
	   XFillArc (theme->wm->dpy, theme->pictures[theme->picture_cnt]->pxmp,
		     theme->gc, x, y, 2*r, 2*r, 0, 360*64);
	 else
	   XDrawArc (theme->wm->dpy, theme->pictures[theme->picture_cnt]->pxmp,
		     theme->gc, x, y, 2*r, 2*r, 0, 360*64);

	 XSetForeground(theme->wm->dpy, theme->mask_gc,
			WhitePixel(theme->wm->dpy, theme->wm->screen));
	 XSetLineAttributes(theme->wm->dpy, theme->mask_gc, lw, LineSolid,
			    CapNotLast, JoinBevel);

	 if (is_filled)
	   XFillArc (theme->wm->dpy, theme->pictures[theme->picture_cnt]->mask,
		     theme->gc, x, y, 2*r, 2*r, 0, 360*64);
	 else
	   XDrawArc (theme->wm->dpy, theme->pictures[theme->picture_cnt]->mask,
		     theme->gc, x, y, 2*r, 2*r, 0, 360*64);
	 continue;
      }
      if (!(strcmp("box", inode->tag)))
      {
	 int x,y,w,h,lw;
	 char *color;
	 Tcolor *tcol;
	 Bool is_filled = False;
	 
	 GET_INT_ATTR(inode,"x", x);
	 GET_INT_ATTR(inode,"y", y);
	 GET_INT_ATTR(inode,"w", w);
	 GET_INT_ATTR(inode,"h", h);
	 lw = (get_attr(inode, "width") ? atoi(get_attr(inode, "width")) : 1);
	 if ( (color = get_attr(inode, "color")) == NULL)
	 { return 0; }
	 if ((tcol = tcolor_lookup(theme, color)) == NULL)
	    return 0;
	 if (get_attr(inode, "options")
	     && (!strcmp("filled", get_attr(inode, "options"))))
	    is_filled = True;
	 
	 XSetForeground(theme->wm->dpy, theme->gc, tcol->xcol.pixel);
	 XSetLineAttributes(theme->wm->dpy, theme->gc, lw, LineSolid,
			    CapNotLast, JoinMiter);

	 if (is_filled)
	    XFillRectangle(theme->wm->dpy,
			   theme->pictures[theme->picture_cnt]->pxmp,
			   theme->gc, x, y, w, h);
	 else
	    XDrawRectangle(theme->wm->dpy,
			   theme->pictures[theme->picture_cnt]->pxmp,
			   theme->gc, x, y, w, h);

	 XSetForeground(theme->wm->dpy, theme->mask_gc,
			WhitePixel(theme->wm->dpy, theme->wm->screen));
	 XSetLineAttributes(theme->wm->dpy, theme->mask_gc, lw, LineSolid,
			    CapNotLast, JoinBevel);

	 if (is_filled)
	    XFillRectangle(theme->wm->dpy,
			   theme->pictures[theme->picture_cnt]->mask,
			   theme->mask_gc, x, y, w, h);
	 else
	    XDrawRectangle(theme->wm->dpy,
			   theme->pictures[theme->picture_cnt]->mask,
			   theme->mask_gc, x, y, w, h);
	 continue;
      }

      /* todo ovals. filled ovals, boxs, filled boxs */
      /* remember to create a mask too! */
   }
   theme->picture_cnt++;
   return 1;
}

Tlabel *tlabel_new(MBtheme *t, Tfont *font,
		   Tcolor *col , Tcolor *bgcol, char *justify)
{
   Tlabel *l = (Tlabel *)malloc(sizeof(Tlabel));

   if (!strcmp(justify, "left"))
   { l->justify = left; }
   else if (!strcmp(justify, "center"))
   { l->justify = center; }
   else if (!strcmp(justify, "right"))
   { l->justify = right; }
   else
   { l->justify = left; }

   l->col = col;
   l->bgcol = bgcol;
   l->font = font;
   
   return l;
}

Tbutton *tbutton_new(MBtheme *t)
{
   Tbutton *b = (Tbutton *)malloc(sizeof(Tbutton));
   return b;
}

static layer_type
parse_layer_type(char *def_str)
{
   if (def_str == NULL) return plain; /* defualt  */
   if (!strcmp(def_str, "plain"))
   { return plain; }
   else if (!strcmp(def_str, "gradient-horiz"))
   { return gradientHoriz; }
   else if (!strcmp(def_str, "gradient-vert"))
   { return gradientVert; }
   else if (!strcmp(def_str, "pixmap"))
   { return pixmap; }
   else if (!strcmp(def_str, "pixmap-tiled"))
   { return pixmapTiled; }
   else if (!strcmp(def_str, "picture"))
   { return picture; }
   else if (!strcmp(def_str, "label"))
   { return label; }
   else if (!strcmp(def_str, "icon"))
   { return icon; }
   else { return plain; }
}

static int
parse_button_action(char *def_str)
{
   if (def_str == NULL) return 0; /* defualt  */
   if (!strcmp(def_str, "next"))
   { return ACTION_NEXT; }
   else if (!strcmp(def_str, "prev"))
   { return ACTION_PREV; }
   else if (!strcmp(def_str, "close"))
   { return ACTION_CLOSE; }
   else if (!strcmp(def_str, "minimize"))
   { return ACTION_MIN; }
   else if (!strcmp(def_str, "maximize"))
   { return ACTION_MAX; }
   else if (!strcmp(def_str, "menu"))
   { return ACTION_MENU; }
   else if (!strcmp(def_str, "hide"))
   { return ACTION_HIDE; }
   else { return 0; }
}

static int
parse_frame_id(char *def_str)
{
   if (def_str == NULL) return 0; /* defualt  */
   if (!strcmp(def_str, "main"))
   { return FRAME_MAIN; }
   else if (!strcmp(def_str, "dialog"))
   { return FRAME_DIALOG; }
   else if (!strcmp(def_str, "utility-min"))
   { return FRAME_UTILITY_MIN; }
   else if (!strcmp(def_str, "utility-max"))
   { return FRAME_UTILITY_MAX; }
   else if (!strcmp(def_str, "menu"))
   { return FRAME_MENU; }
   else { return 0; }
}


static int
handle_frame_tag(MBtheme *theme, XMLNode *node)
{
   Nlist *n;
   XMLNode *inode;
   int current_layer  = 0;
   int current_button = 0;
   int i;
   
   for(i=0;i<MAX_BUTTONS;i++)
      theme->frames[theme->frame_cnt].buttons[i] = NULL;
   
   if ( !(theme->frames[theme->frame_cnt].id
	 = parse_frame_id(get_attr(node, "id"))) )
   { printf("couldn't get node ID %s \n", get_attr(node, "id")); return 0; }

   GET_INT_ATTR(node,"size", theme->frames[theme->frame_cnt].size);

   theme->frames[theme->frame_cnt].option_str = get_attr(node, "options");

   if (theme->frames[theme->frame_cnt].id == FRAME_MENU
       && theme->frames[theme->frame_cnt].option_str != NULL
       && (strstr(theme->frames[theme->frame_cnt].option_str, "noicons")))
     {
       theme->wm->config->use_icons = False;
     }
   
   for(n = node->kids; n != NULL; n = n->next)
   {
      inode = n->data;
      if (!(strcmp("layer", inode->tag)))
      {

	 Tlayer *tl = &theme->frames[theme->frame_cnt].layers[current_layer];
	 
	 if ( (tl->x = gparam_new(get_attr(inode, "x"))) == NULL)
	    return 0;
	 if ( (tl->y = gparam_new(get_attr(inode, "y"))) == NULL)
	    return 0;
	 if ( (tl->w = gparam_new(get_attr(inode, "w"))) == NULL)
	    return 0;
	 if ( (tl->h = gparam_new(get_attr(inode, "h"))) == NULL)
	    return 0;
	 tl->type = parse_layer_type(get_attr(inode, "type"));

	 switch (tl->type)
	 {
	    case plain:
	       tl->color = tcolor_lookup(theme, get_attr(inode, "color"));
	       break;
	    case gradientHoriz:
	    case gradientVert:
	    case pixmap:
	    case pixmapTiled:
	       tl->tpxmp = tpixmap_lookup(theme,
					   get_attr(inode, "pixmap"));
	       break;
	    case picture:
	       tl->picture = tpicture_lookup(theme,
					     get_attr(inode, "picture"));
	       break;
	    case label:
	       tl->label = tlabel_new(theme,
				      tfont_lookup(theme,
						    get_attr(inode, "font")),
				      tcolor_lookup(theme,
						    get_attr(inode, "color")),
				      tcolor_lookup(theme,
						    get_attr(inode,"bgcolor")),
				      get_attr(inode, "justify"));
	       break;
	    case icon:
	       break; 	/* noting to do here  */
	 }
	 current_layer++;
	 continue;
      }
      
      if (!(strcmp("button", inode->tag)))
      {
	 Tbutton *tb;
	 int action = parse_button_action(get_attr(inode, "action"));

	 if (!action) return 0;
	 
	 theme->frames[theme->frame_cnt].buttons[action]
	    = tbutton_new(theme);
	 tb = theme->frames[theme->frame_cnt].buttons[action];

	 tb->wants_dbl_click = False;
	 tb->inputonly       = False;

	 /* TODO: should allow defualts for this espcially type != inputonly */
	 if ( (tb->x = gparam_new(get_attr(inode, "x"))) == NULL)
	    return 0;
	 if ( (tb->y = gparam_new(get_attr(inode, "y"))) == NULL)
	    return 0;
	 if ( (tb->w = gparam_new(get_attr(inode, "w"))) == NULL)
	    return 0;
	 if ( (tb->h = gparam_new(get_attr(inode, "h"))) == NULL)
	    return 0;
	 tb->action = action;

	 
	 if (get_attr(inode, "type"))
	 {
	    if (!strcmp(get_attr(inode, "type"), "inputonly"))
	    {
	       tb->inputonly = True;
	        if (get_attr(inode, "doubleclick"))
		   tb->wants_dbl_click = True;
	    }
	 }
	 
	 tb->active = NULL; tb->inactive = NULL;
	 
	 if (inode->kids) {
	    Nlist *nn;
	    XMLNode *iinode;
	    for(nn = inode->kids; nn != NULL; nn = nn->next)
	    {
	       iinode = nn->data;
	       if (!(strcmp("active", iinode->tag)))
	       {
		  dbg("loaded active button\n");
		  if (get_attr(iinode, "pixmap"))
		     tb->active = tpixmap_lookup(theme,
						 get_attr(iinode, "pixmap"));
		  if (get_attr(iinode, "picture"))
		     tb->active = tpicture_lookup(theme,
						 get_attr(iinode, "picture"));
		  if (tb->active == NULL)
		  {
		     fprintf(stderr, 
			     "could not find pict or pix for active button\n");
		     exit(0);
		  }
		  continue;
	       }
	       if (!(strcmp("inactive", iinode->tag)))
	       {
		  dbg("loaded inactive button\n");
		  if (get_attr(iinode, "pixmap"))
		     tb->inactive = tpixmap_lookup(theme,
						 get_attr(iinode, "pixmap"));
		  if (get_attr(iinode, "picture"))
		     tb->inactive = tpicture_lookup(theme,
						 get_attr(iinode, "picture"));
		  if (tb->inactive == NULL)
		  {
		     printf("could not find pict or pix for inactive button\n");
		     exit(0);
		  }
		  continue;
	       }
	    }
	    if (tb->active == NULL) tb->active = tb->inactive;
	 }
	 
	 current_button++;
	 continue;
      }
      
   }
   theme->frames[theme->frame_cnt].layer_cnt = current_layer;
   theme->frames[theme->frame_cnt].button_cnt = current_button;
   theme->frame_cnt++;
   return 1;
}

static int
handle_global_tag(MBtheme *theme, XMLNode *node)
{
   char *tmp;
   if ( (tmp = get_attr(node, "borderWidth")) == NULL)
   { return 0; }
   
   theme->wbw = strtol(tmp, (char **)NULL, 10);
   
   if ( (tmp = get_attr(node, "borderColor")) == NULL)
   { return 0; }

   theme->wbc = tcolor_lookup(theme, tmp);

   if ( (tmp = get_attr(node, "font")) == NULL)
   { return 0; }
   
   theme->default_font = tfont_lookup(theme, get_attr(node, "font"));
   
   if ( (tmp = get_attr(node, "fontColor")) == NULL)
   { return 0; }
    
   theme->default_font_color
      = tcolor_lookup(theme, get_attr(node, "fontColor"));
   
   return 1;
}

MBtheme *
mbtheme_new(Wm *w)
{
   XGCValues gv;
   MBtheme *t = (MBtheme *)malloc(sizeof(MBtheme));

   /*
   if (!(t->font = XLoadQueryFont(w->dpy, "fixed")))
   {
      printf("fixed font not found, giving up...\n");
      exit(1);
   }
   
   gv.font    = t->font->fid;   
   */
   gv.graphics_exposures = False;
   gv.function   = GXcopy;
   t->gc = XCreateGC(w->dpy, w->root,
		     GCGraphicsExposures|GCFunction/*|GCFont*/, &gv);

   gv.function = GXinvert;
   gv.subwindow_mode = IncludeInferiors;
   gv.line_width = 1;
   t->band_gc = XCreateGC(w->dpy, w->root,
			  GCFunction|GCSubwindowMode|GCLineWidth, &gv);
   
   t->mask_gc = None;
   t->wm = w;
   t->color_cnt = 0;
   
   return t;
}

#ifndef STANDALONE

static void
mbtheme_save_selected(MBtheme *t, char *str)
{
  FILE* fp;
  char filename[256];
  
  strcpy(filename,(char *)getenv("HOME"));
  strcat(filename,"/.mbtheme");
  dbg("writing to %s\n", filename);
  if (!(fp = fopen(filename, "w"))) return;
  fprintf(fp,"%s", str);
  fclose(fp);
}

static char *
mbtheme_load_selected(void)
{
  struct stat st;
  FILE* fp;
  char* str;
  int len;
  char filename[256];
  
  strcpy(filename,(char *)getenv("HOME"));
  strcat(filename,"/.mbtheme");

  dbg("about to open %s\n", filename);
  if (stat(filename, &st)) return NULL;
  
  if (!(fp = fopen(filename, "r"))) return NULL;

  /* Read in the file. */
  str = (char *)malloc(sizeof(char)*(st.st_size + 1));
  len = fread(str, 1, st.st_size, fp);
  if (len >= 0) str[len] = '\0';
  
  fclose(fp);
  
  return str;
}


void
mbtheme_switch(Wm *w, char *new_theme_conf)
{
   /* -- save theme in Xdefaults -- broke ?
   static XrmDatabase rdb;
   char *enviroment;
   XrmValue v;
   
   if ((enviroment = getenv("XENVIROMENT")) == NULL)
   {
      enviroment = (char *)getenv("HOME");
      strcat(enviroment, "/.Xdefaults");
   }
   rdb = XrmGetFileDatabase(enviroment);
   v.size = sizeof(char)*strlen(new_theme_conf);
   v.addr = new_theme_conf;
   XrmPutResource(&rdb, "matchbox.theme", "string", &v);
   */
  mbtheme_save_selected(w->mbtheme, new_theme_conf);
  execlp("matchbox", "matchbox", "-theme", new_theme_conf, 0);
  err("couldn't exec, is matchbox in your $PATH ?");
}

#endif


void
mbtheme_init(Wm *w, char *theme_conf)
{
   XMLNode *root_node, *cnode;
   Nlist *n;
   char orig_wd[255];
   XMLParser *parser = new_xmlparser();

#ifdef STANDALONE

#include "standalone.h"

   root_node = parse_data_dom(parser, xmldata);
   w->mbtheme = mbtheme_new(w);

#else

   char *default_theme_conf = DEFAULTTHEME ;

   if (theme_conf == NULL) { 
     theme_conf = mbtheme_load_selected();
     if (theme_conf == NULL) { 
       theme_conf = default_theme_conf;
     } 
   }

   XChangeProperty(w->dpy, w->root, w->mb_theme, XA_STRING, 8,
		   PropModeReplace, theme_conf, strlen(theme_conf));
   

   w->mbtheme = mbtheme_new(w);

   if (getcwd(orig_wd, 255) == (char *)NULL)
   {
      printf("Cant get current directory\n");
      exit(0);
   }

   chdir(theme_conf);
   
   root_node = parse_file_dom(parser, "theme.xml");

#endif

   if (root_node == NULL)
   {
      printf("failed to parse theme file \n");
      exit(0);
   }
   
   w->mbtheme->frame_cnt = 0;
   
   for(n = root_node->kids; n != NULL; n = n->next)
   {
      cnode = n->data;
      if (!strcmp("color",cnode->tag))
      {
	 if (!handle_color_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad picture tag\n");
	    exit(0);
	 }
	 continue;
      }
      if (!strcmp("font",cnode->tag))
      {
	 if (!handle_font_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad font tag\n");
	    exit(0);
	 }
	 continue;
      }
      if (!strcmp("border",cnode->tag))
      {
	 if (!get_attr(cnode, "width"))
	 {
	    fprintf(stderr, "bad border tag\n");
	    exit(0);
	 }
	 w->mbtheme->wbw = atoi(get_attr(cnode, "width"));
      }

      if (!strcmp("frame",cnode->tag))
      {
	 if (!handle_frame_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad frame tag\n");
	    exit(0);
	 }
	 continue;
      }
      if (!strcmp("picture",cnode->tag))
      {
	 if (!handle_picture_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad picture tag\n");
	    exit(0);
	 }
	 continue;
      }

      if (!strcmp("pixmap",cnode->tag))
      {
	 if (!handle_pixmap_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad picture tag\n");
	    exit(0);
	 }
	 continue;
      }

      if (!strcmp("global",cnode->tag))
      {
	 dbg("got global tag");
	 if (!handle_global_tag(w->mbtheme, cnode))
	 {
	    fprintf(stderr, "bad global tag\n");
	    exit(0);
	 }
	 continue;
      }

      fprintf(stderr, "Dont know what to do with a  %s tag\n", cnode->tag);
   }

   chdir(orig_wd);
   
   free_xmlparser(parser, root_node);  /* TODO: make it actually free better  */
   
}


