/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "misc.h"

void err(const char *fmt, ...)
{
#ifdef DEBUG
    va_list argp;
    fprintf(stderr, "X-error: ");
    va_start(argp, fmt);
    vfprintf(stderr, fmt, argp);
    va_end(argp);
    fprintf(stderr, "\n");
#endif
}

/*
void dbg(const char *fmt, ...)
{
    va_list argp;
    fprintf(stderr, "DBG: ");
    va_start(argp, fmt);
    vfprintf(stderr, fmt, argp);
    va_end(argp);
    fprintf(stderr, "\n");
}
*/

int max(int n, ...)
{
   va_list ap;
   int maxnum = n;

   va_start(ap, n);
   while ((n = va_arg(ap, int)))
   {
      if (n > maxnum)
	 maxnum = n;

   }
   va_end(ap);
   return maxnum;
}



void fork_exec(char *cmd)
{
    pid_t pid = fork();

    switch (pid) {
        case 0:
            execlp("/bin/sh", "sh", "-c", cmd, NULL);
            err("exec failed, cleaning up child");
            exit(1);
        case -1:
            err("can't fork");
    }
}

void sig_handler(int signal)
{
    switch (signal) {
        case SIGINT:
        case SIGTERM:
        case SIGHUP:
	   /* quit_nicely(); */ exit(1); break;
        case SIGCHLD:
            wait(NULL); break;
    }
}

int handle_xerror(Display *dpy, XErrorEvent *e)
{
   //    client *c = wm_find_client(w, e->resourceid, WINDOW);

    if (e->error_code == BadAccess &&
	e->resourceid == RootWindow(dpy, DefaultScreen(dpy)) ) {
        err("root window unavailible (maybe another wm is running?)");
        exit(1);
    } else {
        char msg[255];
        XGetErrorText(dpy, e->error_code, msg, sizeof msg);
        err("X error (%#lx): %s (opcode: %i)",
	    e->resourceid, msg, e->request_code);
    }

    /* if (c) remove_client(c, WITHDRAW); */
    return 0;
}

int ignore_xerror(Display *dpy, XErrorEvent *e)
{
   return 0;
}
void dump(client *c)
{
    err("%s\n\t ignore %d\n"
        "\tframe %#lx, win %#lx, geom %dx%d+%d+%d",
        c->name, c->ignore_unmap,
        c->frame, c->window, c->width, c->height, c->x, c->y);
}

void dump_clients(wm* w)
{
    client *c;
    for (c = w->head_client; c; c = c->next) dump(c);
}

void
scale_icon(Client *c)  /* Stolen from fspanel  */
{
   int xx, yy, x, y, w, h, d, bw;
   Pixmap pix, mk = None;
   XGCValues gcv;
   GC mgc = None;
   int size = c->wm->config->use_icons;
   
   XGetGeometry (c->wm->dpy, c->icon, &pix, &x, &y, &w, &h, &bw, &d);
   pix = XCreatePixmap (c->wm->dpy, c->wm->root,
			size, size,
			DefaultDepth(c->wm->dpy, c->wm->screen));

   if (c->icon_mask)
   {
      mk = XCreatePixmap (c->wm->dpy, c->wm->root, size, size, 1);
      gcv.subwindow_mode = IncludeInferiors;
      gcv.graphics_exposures = False;
      mgc = XCreateGC(c->wm->dpy, mk,
		      GCGraphicsExposures|GCSubwindowMode, &gcv);
   }

   /* this is my simple & dirty scaling routine */
   for (y = size - 1; y >= 0; y--)
   {
      yy = (y * h) / size;
      for (x = size - 1; x >= 0; x--)
      {
	 xx = (x * w) / size;
	 if (d != DefaultDepth(c->wm->dpy, c->wm->screen))
	 {
	    XCopyPlane (c->wm->dpy, c->icon, pix, c->wm->mbtheme->gc,
			xx, yy, 1, 1, x, y, 1);
	 }
	 else
	 {
	    XCopyArea (c->wm->dpy, c->icon, pix, c->wm->mbtheme->gc,
		       xx, yy, 1, 1, x, y);
	 }
	 if (mk != None)
	    XCopyArea (c->wm->dpy, c->icon_mask, mk, mgc,
		       xx, yy, 1, 1, x, y);
      }
   }
   
   if (mk != None)
   {
      XFreeGC (c->wm->dpy, mgc);
      c->icon_mask = mk;
   }

   c->icon = pix;
}

