/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "select_client.h"

client*
select_client_new(Wm *w)
{
   Window win;
   XSetWindowAttributes attr;
   Tframe *frm = theme_get_frame(w->mbtheme, FRAME_MAIN );
   int width,height;
   Client *c;
   
   select_client_calc_dim(w,&width,&height);

   if (width == 0 || height ==0) return NULL;

   if (XGrabPointer(w->dpy, w->root, True,
		    (ButtonPressMask|ButtonReleaseMask),
		    GrabModeAsync,
		    GrabModeAsync, None, w->curs, CurrentTime)
       != GrabSuccess)
      return NULL;

   
   attr.override_redirect = True;
   attr.background_pixel = w->mbtheme->wbc->xcol.pixel;
   attr.border_pixel = w->mbtheme->wbc->xcol.pixel;
   attr.event_mask = ButtonPressMask|ExposureMask|EnterWindowMask;

   win = XCreateWindow(w->dpy, w->root,
		       gparam_convert(frm, frm->buttons[ACTION_MENU]->x,
				      HORIZONTAL),
		       main_client_title_height(w->main_client),
		       width, height, WBW(w->main_client),
		       CopyFromParent, CopyFromParent, CopyFromParent,
		       CWOverrideRedirect|CWBackPixel|CWBorderPixel|CWEventMask,
		       &attr);


   
   c = base_client_new(w, win);
   dbg("SELECT: win craeted with id %li", win);
   c->type = menu;
   c->title_frame = c->frame = c->window;
   
   c->destroy      = &select_client_destroy;
   c->button_press = &select_client_button_press;
   c->redraw       = &select_client_redraw;

   w->flags ^= MENU_FLAG; /* turn on menu flag */

   c->redraw(c, False);
   c->mapped = True;
   client_set_state(c, NormalState);
   
   XMapWindow(c->wm->dpy, c->frame);

   return c;
}

void
select_client_calc_dim(Wm *w, int *width, int *height)
{
   int twidth = 0, theight = 0;
   int tmp;
   client *p;
   MBtheme *t = w->mbtheme;
   int j;
   int space_avail = w->dpy_width - 16;

   START_CLIENT_LOOP(w,p);   
      if (p->type == mainwin && p->name
	  && p != w->main_client 
	  && client_get_state(p) == NormalState)
      {

	 j = strlen(p->name);

	 while ((tmp = theme_get_text_length(t, t->default_font, p->name, j))
		> space_avail && j>0)
	   j--;

	 tmp += ENTRY_SPC;

         theight += MAX( w->config->use_icons,
			 t->default_font->font->ascent
			 + t->default_font->font->descent + ENTRY_SPC );
	 tmp += w->config->use_icons;
	 if (tmp > twidth) twidth = tmp;
      }
    END_CLIENT_LOOP(w,p);

    twidth += ENTRY_SPC;

    *width = twidth;
    *height = theight;
}

void
select_client_button_press(Client *c, XButtonEvent *e)
{
   client *p;
   int y = 0;
   MBtheme *t = c->wm->mbtheme;

    int font_height = MAX( c->wm->config->use_icons,
			   t->default_font->font->ascent
			   +t->default_font->font->descent +
			   ENTRY_SPC );

   for(p = c->prev; p != c ; p = p->prev)
      if (p->type == mainwin && p->name
	  && p != c->wm->main_client 
	  && client_get_state(p) == NormalState )
      {
	 if (e->y > y &&
	     e->y < (y+font_height))
	 {
	    c->wm->main_client->hide(c->wm->main_client);
	    p->show(p);
	    select_client_destroy(c);
	    return;
	 }
	 y += font_height;
      }
   select_client_destroy(c);
}

void
select_client_destroy(Client *c)
{
   dbg("****select window destroyed****\n");
   XUngrabPointer(c->wm->dpy, CurrentTime);
   c->wm->flags ^= MENU_FLAG;
   XUnmapWindow(c->wm->dpy, c->frame);
   base_client_destroy(c);
}

void
select_client_redraw(client *c, Bool use_cache)
{
   client *p;
   MBtheme *t = c->wm->mbtheme;
   int space_avail = c->wm->dpy_width -16;
   int txt_y, y = 0, j = 0;
   Tframe *frm = theme_get_frame(c->wm->mbtheme, FRAME_MENU );

   int h = MAX( c->wm->config->use_icons,
		t->default_font->font->ascent 
		+ t->default_font->font->descent + ENTRY_SPC );
   int font_y = t->default_font->font->ascent;

   if (c->wm->config->use_icons)
      txt_y = (c->wm->config->use_icons + font_y)/2;
   else {
      h += ENTRY_SPC;
      txt_y = font_y;
   }
	 
   if (c->backing == (Pixmap)NULL)
      client_init_backing(c, c->width, c->height);
   
   frm->size = c->height;
   theme_paint( c->wm->mbtheme, c, FRAME_MENU );
   
   for(p = c->prev; p != c ; p = p->prev)
      if (p->type == mainwin && p->name
	  && p != c->wm->main_client 
	  && client_get_state(p) == NormalState )
      {

	 if (p->icon)
	 {
	    XGCValues gc_vals;
	    gc_vals.clip_x_origin = ENTRY_SPC/2;
	    gc_vals.clip_y_origin = y;
	    XChangeGC(c->wm->dpy, c->wm->mbtheme->gc,
		      GCClipXOrigin|GCClipYOrigin, &gc_vals);
	    
	    if (p->icon_mask)
	       XSetClipMask(c->wm->dpy, c->wm->mbtheme->gc, p->icon_mask);
	    
	    XCopyArea(c->wm->dpy, p->icon, c->backing, c->wm->mbtheme->gc, 
		      0, 0, 16, 16, ENTRY_SPC/2, y);
	 }
	 XSetClipMask(c->wm->dpy, c->wm->mbtheme->gc, None);


	 j = strlen(p->name);
	 
	 while (theme_get_text_length(t, t->default_font, p->name, j)
		> space_avail && j > 0)
	   j--;
	 
#ifdef USE_XFT
      XftDrawString8(c->xftdraw, &(t->default_font_color->xftcol), 
		     t->default_font->font,
		     ENTRY_SPC + c->wm->config->use_icons,
		     y + txt_y,
		     (unsigned char *) p->name, j);
#else
      XSetFont(t->wm->dpy, t->gc, t->default_font->font->fid);
      XSetForeground(t->wm->dpy, t->gc, t->default_font_color->xcol.pixel);
      XDrawString(c->wm->dpy, c->backing, t->gc,
		  ENTRY_SPC  + c->wm->config->use_icons,
		  y + txt_y,
		  p->name, j);
#endif
	y += h;
      }

   XSetWindowBackgroundPixmap(c->wm->dpy, c->frame, c->backing);
   XClearWindow(c->wm->dpy, c->frame);


}



