/* miniapm - A tiny battery monitor

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <apm.h>

#define TIME_LEFT  0
#define PERCENTAGE 1
#define AC_POWER   2

#define WIN_WIDTH  8
#define WIN_HEIGHT 14

#define CLOCK_DISP 1
#define BATT_DISP  0

#ifndef AC_LINE_STATUS_ON
#define AC_LINE_STATUS_ON 1
#endif

#include "tray.h"
#include "battery.xpm"

#define RED     0
#define ORANGE  1
#define GREEN   2
#define GRAY    3
#define YELLOW  4

Display* display;	
Window win;
Pixmap mask, icon;
GC gc, rev_gc, mask_gc, draw_gc;
XpmAttributes xpm_attr;	
int apm_vals[3];

XColor Cols[5][4];
XColor grey_col;

int
xcol_from_str(Display *display, XColor *col, const char *defstr)
{
  char *str;
  const char delim[] = ",:";
  char *token;
  XColor exact;

  if ((strchr(defstr, delim[0]) != NULL)
      || (strchr(defstr, delim[1]) != NULL) )
  {
     str = strdup(defstr);

     token = strsep (&str, delim); 
     col->red = ( atoi(token) * 65535 ) / 255; 
     token = strsep (&str, delim); 
     col->green = ( atoi(token) * 65535 ) / 255;
     token = strsep (&str, delim); 
     col->blue = ( atoi(token) * 65535 ) / 255;

     return XAllocColor(display,
			DefaultColormap(display, DefaultScreen(display)),
			col);
  } else {
     return XAllocNamedColor(display,
			     DefaultColormap(display, DefaultScreen(display)),
			     defstr, col, &exact);
  }
}

int read_apm(int *values)
{
  /* add stat function here ? */
 
  apm_info info;
  apm_read(&info);

  values[TIME_LEFT] = info.battery_time;
  values[PERCENTAGE] = info.battery_percentage;
  values[AC_POWER] = info.ac_line_status;

  return 0;
}

void bat_redraw()
{
  int col_ref=0, pwr;
  int ac_power = 0;

  XCopyArea(display, icon, win, gc, 0, 0,
	    xpm_attr.width, xpm_attr.height, 0, 0);

  read_apm(apm_vals);

  //apm_vals[PERCENTAGE] = 50;

  if (apm_vals[PERCENTAGE] <= 0 || apm_vals[PERCENTAGE] > 99)
    { col_ref = GREEN; ac_power = 1; apm_vals[PERCENTAGE] = 100;}
  else if (apm_vals[PERCENTAGE] <= 30)
    col_ref = RED;
  else if (apm_vals[PERCENTAGE] < 75)
    col_ref = ORANGE;
  else if (apm_vals[PERCENTAGE] <= 100)
    col_ref = GREEN;

  if (apm_vals[AC_POWER] == AC_LINE_STATUS_ON) ac_power = 1;
  
  pwr = apm_vals[PERCENTAGE]/10;

  if (pwr)
    {
      XSetForeground(display, draw_gc, Cols[col_ref][0].pixel);
      XDrawLine(display, win, draw_gc, 1, 13-pwr, 1, 11);
      
      XSetForeground(display, draw_gc, Cols[col_ref][1].pixel);
      XDrawLine(display, win, draw_gc, 2, 13-pwr, 2, 12);
      
      XSetForeground(display, draw_gc, Cols[col_ref][2].pixel);
      XDrawLine(display, win, draw_gc, 3, 13-pwr, 3, 12);
      
      XSetForeground(display, draw_gc, Cols[col_ref][1].pixel);
      XDrawLine(display, win, draw_gc, 4, 13-pwr, 4, 12);
      
      XSetForeground(display, draw_gc, Cols[col_ref][0].pixel);
      XDrawLine(display, win, draw_gc, 5, 13-pwr, 5, 11);
    }

  if (pwr <= 9)
    {
      XSetForeground(display, draw_gc, Cols[GRAY][0].pixel);
      XDrawLine(display, win, draw_gc, 1, 3, 1, 12-pwr);
      
      XSetForeground(display, draw_gc, Cols[GRAY][1].pixel);
      XDrawLine(display, win, draw_gc, 2, 3, 2, 12-pwr);
      
      XSetForeground(display, draw_gc, Cols[GRAY][2].pixel);
      XDrawLine(display, win, draw_gc, 3, 3, 3, 12-pwr);
      
      XSetForeground(display, draw_gc, Cols[GRAY][1].pixel);
      XDrawLine(display, win, draw_gc, 4, 3, 4, 12-pwr);
      
      XSetForeground(display, draw_gc, Cols[GRAY][0].pixel);
      XDrawLine(display, win, draw_gc, 5, 3, 5, 12-pwr);
    }

  if (ac_power)
    {
      XSetForeground(display, gc, Cols[YELLOW][0].pixel);
      XDrawPoint(display, win, gc, 0, 4);
      XDrawPoint(display, win, gc, 0, 5);

      XFillRectangle(display, win, gc, 1, 5, 2, 3 );
      XFillRectangle(display, win, gc, 3, 7, 2, 3 );

      XDrawPoint(display, win, gc, 5, 9);
      XDrawPoint(display, win, gc, 5, 10);
    }

}

void printversion(void)
{
  printf("Version: 0.3 \n");
}

void usage(void)
{
  printf("Usage: miniapm <options>\n");
  printf("Options:\n");
  printf("  -display  <display>\n");
  printf("  -v        version\n");
}


int main(int argc, char* argv[])
{
  int screen_num;	


  XGCValues gv;

  Atom window_type_atom;
  Atom window_type_dock_atom;
  Atom mb_align_atom;
  Atom mb_align_east_atom;
  
  char *display_name = (char *)getenv("DISPLAY");  
  int depth, i;

  time_t last_time = 0;
  time_t curTime;

  int update_req = 0;
  int align_right = 0;

  /* pass command line */
  for (i=1; argv[i]; i++) {
    char *arg = argv[i];
    if (*arg=='-') {
      switch (arg[1]) {
	case 'd' : /* display */
	  display_name = argv[i+1];
	  i++;
	  break;
	case 'r' :
	  align_right = 1;
	  break;
	case 'v' :
	  printversion();
	  exit(0);
	  break;
	default:
	  usage();
	  exit(0);
	  break;
      }
    }
  }

  display = XOpenDisplay(display_name);
  if (display == NULL) {
    fprintf(stderr, "%s: cannot connect to X server '%s'\n",
            argv[0], display_name);
    exit(1);
  }

  window_type_atom = XInternAtom(display,
				 "_NET_WM_WINDOW_TYPE" , False);
  window_type_dock_atom = XInternAtom(display,
				      "_NET_WM_WINDOW_TYPE_DOCK",False);
  
  mb_align_atom
    = XInternAtom(display, "_MB_DOCK_ALIGN", False);
  mb_align_east_atom
    = XInternAtom(display, "_MB_DOCK_ALIGN_EAST", False);

  screen_num = DefaultScreen(display);
  win = XCreateSimpleWindow(display,
			    RootWindow(display, screen_num),
			    0, 0,
			    8, 15,
			    0, BlackPixel(display, screen_num),
			    WhitePixel(display, screen_num));
  
  gv.graphics_exposures = False;
  gv.function   = GXcopy;
  gv.foreground = BlackPixel(display, screen_num);
  
  gc = XCreateGC(display, RootWindow(display, screen_num),
		 GCGraphicsExposures|GCFunction|GCForeground, &gv);
  
  gv.foreground = WhitePixel(display, screen_num); 
  draw_gc = XCreateGC(display, RootWindow(display, screen_num),
		 GCGraphicsExposures|GCFunction|GCForeground, &gv);

  xcol_from_str(display, &(Cols[RED][0]), "100:0:0");
  xcol_from_str(display, &(Cols[RED][1]), "200:0:0");
  xcol_from_str(display, &(Cols[RED][2]), "255:0:0");

  xcol_from_str(display, &Cols[ORANGE][0], "255:100:0");
  xcol_from_str(display, &Cols[ORANGE][1], "255:200:100");
  xcol_from_str(display, &Cols[ORANGE][2], "255:255:150");

  xcol_from_str(display, &Cols[GREEN][0], "0:100:0");
  xcol_from_str(display, &Cols[GREEN][1], "0:200:0");
  xcol_from_str(display, &Cols[GREEN][2], "0:255:0");

  xcol_from_str(display, &Cols[GRAY][0], "100:100:100");
  xcol_from_str(display, &Cols[GRAY][1], "200:200:200");
  xcol_from_str(display, &Cols[GRAY][2], "255:255:255");

  xcol_from_str(display, &Cols[YELLOW][0], "255:255:100");

  depth = DefaultDepth(display, DefaultScreen(display));

  xpm_attr.valuemask = XpmColormap | XpmCloseness;
  xpm_attr.closeness = 40000;
  xpm_attr.colormap = DefaultColormap(display, DefaultScreen(display));

  if (XpmCreatePixmapFromData( display, win, battery_xpm,
			       &icon, &mask, &xpm_attr) != XpmSuccess )
    {
      fprintf(stderr, "failed to get icon image\n");
      exit(1);
    }
  
  XCopyArea(display, icon, win, gc, 0, 0,
	    xpm_attr.width, xpm_attr.height, 0, 0);
  XShapeCombineMask (display, win, ShapeBounding, 0, 0, mask, ShapeSet);
  
  if (align_right)
    XChangeProperty(display, win, mb_align_atom, XA_ATOM, 32, 
		    PropModeReplace,
		    (unsigned char *) &mb_align_east_atom, 1);

  XStoreName(display, win, "Battery Monitor");

  bat_redraw();
  
  tray_init_session_info(display, win, argv, argc);
  tray_init(display, win);

  XSelectInput(display, win, ExposureMask | StructureNotifyMask |
	       SubstructureNotifyMask | ButtonPressMask); 
  {
    int done = 0;
    XEvent an_event;
    char tray_msg[256];
    
    while (!done)
      {
	curTime = time(0);
	if ( curTime != last_time )
	  {
	    last_time = curTime;
	    update_req = 1;
	  }
	while (XPending(display)) {
	  XNextEvent(display, &an_event);
	  switch (an_event.type) {
	    case Expose:
	      update_req = 1;
	      break;
	    case DestroyNotify:	      
	      done = 1;
	      break;
	     case ButtonPress:
		sprintf(tray_msg, "%s\nJuice %i %%\nTime left: %i mins\n",
			(apm_vals[AC_POWER] == AC_LINE_STATUS_ON) ?
			"Power online" : "Power offline",
			apm_vals[PERCENTAGE], apm_vals[TIME_LEFT]);
		tray_send_message(display, win, tray_msg);
	    default:
	      break;
	  }
	  tray_handle_event(display, win, &an_event);
	}
	if (update_req)
	  {
	    bat_redraw();
	    update_req = 0;
	  } 
	usleep(300000L);
      } 
    
  }

  XFreePixmap(display,mask);
  XCloseDisplay(display);
  
  return 0;
}









