/* minitime - a mini dockable clock

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

#include "minitimestrip.xpm"
#include "tray.h"

#define TIME_DISP 1
#define DATE_DISP 0
#define PADDING   2

#define NUM_WIDTH 7

void printversion(void)
{
  printf("Version: 0.2 \n");
}

void usage(void)
{
  printf("Usage: minitime <options>\n");
  printf("Options:\n");
  printf("  -display  <display>\n");
  printf("  -r        align right in dock ( matchbox only )\n");
  printf("  -m        show date rather than time\n");
  printf("  -v        version\n");
}

int main(int argc, char* argv[])
{
  Display* display;	
  int screen_num;	
  Window win;
  GC gc;
  Pixmap mask;
  Pixmap clock_pxmp;
  XpmAttributes xpm_attr;	
  XGCValues gv;
  
  char *display_name = (char *)getenv("DISPLAY");
  int depth, i;
  int mode = TIME_DISP;

  struct timeval  tv;
  struct timezone tz;
  struct tm *localTime = NULL; 
  
  time_t actualTime;
  long   actualMinutes = 0, last_actualMinutes = 0;
  
  int update_req = 0;


  Atom window_type_atom;
  Atom window_type_dock_atom;

  Atom mb_align_atom;
  Atom mb_align_east_atom;
  int align_dock_right = False;
  
  /* pass command line */
  for (i=1; argv[i]; i++) {
    char *arg = argv[i];
    if (*arg=='-') {
       switch (arg[1]) {
	  case 'd' : /* display */
	     display_name = argv[i+1];
	     i++;
	     break;
	  case 'm' :
	     mode = DATE_DISP;
	     break;
	  case 'r' :
	     align_dock_right = True;
	     break;
	  case 'v' :
	     printversion();
	     exit(0);
	     break;
	  default:
	     usage();
	     exit(0);
	     break;
       }
    }
  }
  
  display = XOpenDisplay(display_name);
  if (display == NULL) {
     fprintf(stderr, "%s: cannot connect to X server '%s'\n",
	     argv[0], display_name);
     exit(1);
  }

  window_type_atom
     = XInternAtom(display, "_NET_WM_WINDOW_TYPE", False);
  window_type_dock_atom
     = XInternAtom(display, "_NET_WM_WINDOW_TYPE_DOCK",False);

  mb_align_atom
     = XInternAtom(display, "_MB_DOCK_ALIGN", False);
  mb_align_east_atom
     = XInternAtom(display, "_MB_DOCK_ALIGN_EAST", False);
  
  
  screen_num = DefaultScreen(display);
  
  win = XCreateSimpleWindow(display,
			    RootWindow(display, screen_num),
			    0, 0,
			    38, 11,
			    0, BlackPixel(display, screen_num),
			    WhitePixel(display, screen_num));
  
  gv.graphics_exposures = 0;
  gv.function   = GXcopy;
  gc = XCreateGC(display, RootWindow(display, screen_num),
		 GCGraphicsExposures|GCFunction, &gv);
  
  depth = DefaultDepth(display, DefaultScreen(display));

  xpm_attr.valuemask = XpmColormap | XpmCloseness;
  xpm_attr.closeness = 40000;
  xpm_attr.colormap = DefaultColormap(display, screen_num);

  XCreatePixmapFromData(display, win, minitimestrip_xpm, &clock_pxmp,
			&mask, &xpm_attr);

  if (align_dock_right)
     XChangeProperty(display, win, mb_align_atom, XA_ATOM, 32, 
		     PropModeReplace,
		     (unsigned char *) &mb_align_east_atom, 1);

  XStoreName(display, win, "Clock");

  tray_init_session_info(display, win, argv, argc);
  tray_init(display, win);


  XSelectInput(display, win, ExposureMask | StructureNotifyMask |
	       SubstructureNotifyMask | ButtonPressMask |  
	       ButtonReleaseMask ); 
  {
    int done = 0;
    XEvent an_event;

    while (!done)
      {
	 gettimeofday(&tv, &tz);
	 actualTime = tv.tv_sec;
	 actualMinutes = actualTime / 60;
	 localTime = localtime(&actualTime);
	 
	if ( actualMinutes != last_actualMinutes )
	  {
	    last_actualMinutes = actualMinutes;
	    update_req = 1;
	  }
	while (XPending(display))
	{
	  XNextEvent(display, &an_event);
	  switch (an_event.type) {
	    case Expose:
	      update_req = 1;
	      break;
	    case DestroyNotify:	      
	      done = 1;
	      break;
	    case ButtonPress:
	       tray_send_message(display, win, ctime(&actualTime));
	       /* mode ^= 1; */
	       update_req = 1;
	      break;
	    default:
	      break;
	  }
	  tray_handle_event(display, win, &an_event);
	}
	if (update_req)
	{
	   XCopyArea(display, clock_pxmp, win, gc, 84, 0, 38, 11, 0, 0);
	   if (mode == TIME_DISP)
	   {
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_hour / 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, 0, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_hour % 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, NUM_WIDTH+1, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_min / 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, NUM_WIDTH*3, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_min % 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, (NUM_WIDTH*4)+1, 0);
	   } else {
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_mday / 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, 0, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			(localTime->tm_mday % 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, NUM_WIDTH+1, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			((localTime->tm_mon+1) / 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, NUM_WIDTH*3, 0);
	      XCopyArea(display, clock_pxmp, win, gc,
			((localTime->tm_mon+1) % 10) * NUM_WIDTH, 0,
			NUM_WIDTH, 11, (NUM_WIDTH*4)+1, 0);
	   }	      

	   update_req = 0;
	} 
	usleep(50000L);
      }
  }
  XFreePixmap(display,mask);
  XCloseDisplay(display);
  
  return 0;
}








